from __future__ import annotations

import logging
from collections import defaultdict
from typing import TYPE_CHECKING

from qcelemental.models import FailedOperation
from sqlalchemy import select, delete, update, or_, text
from sqlalchemy.orm import (
    joinedload,
    selectinload,
    defer,
    undefer,
    with_polymorphic,
    aliased,
    load_only,
)

from qcfractal.components.auth.db_models import UserIDMapSubquery
from qcfractal.components.dataset_db_views import DatasetDirectRecordsView
from qcfractal.components.managers.db_models import ComputeManagerORM
from qcfractal.components.record_db_views import RecordDirectChildrenView, RecordChildrenView
from qcfractal.components.services.db_models import ServiceQueueORM, ServiceDependencyORM
from qcfractal.components.tasks.db_models import TaskQueueORM
from qcfractal.db_socket.helpers import (
    get_general,
    delete_general,
)
from qcportal.all_inputs import AllQCPortalInputTypes, AllInputTypes
from qcportal.all_results import AllSchemaV1ResultTypes, AllQCPortalRecordTypes, AllResultTypes
from qcportal.exceptions import UserReportableError, MissingDataError
from qcportal.managers.models import ManagerStatusEnum
from qcportal.metadata_models import DeleteMetadata, UpdateMetadata
from qcportal.record_models import PriorityEnum, RecordStatusEnum, OutputTypeEnum
from qcportal.utils import chunk_iterable, now_at_utc, is_included
from .base_record_socket import BaseRecordSocket
from .outputstore.utils import create_output_orm
from .record_db_models import (
    RecordComputeHistoryORM,
    BaseRecordORM,
    RecordInfoBackupORM,
    RecordCommentORM,
    OutputStoreORM,
    NativeFileORM,
)
from .record_utils import (
    build_extras_properties,
    upsert_output,
    compute_history_orms_from_qcportal_record,
    compute_history_orm_from_schema_v1,
    native_files_orms_from_qcportal_record,
    native_files_orms_from_schema_v1,
)

if TYPE_CHECKING:
    from sqlalchemy.orm.session import Session
    from sqlalchemy.sql import Select
    from qcfractal.db_socket.socket import SQLAlchemySocket
    from qcportal.record_models import RecordQueryFilters
    from typing import List, Dict, Tuple, Optional, Sequence, Any, Iterable, Type, Union


_default_error = {"error_type": "not_supplied", "error_message": "No error message found on task."}


class RecordSocket:
    """
    Root socket for all record sockets

    This socket contains lots of functionality that is common for all records, as well as
    acts as a root socket for all specific record sockets.
    """

    def __init__(self, root_socket: SQLAlchemySocket):
        self.root_socket = root_socket
        self._logger = logging.getLogger(__name__)

        # All the subsockets
        from .services.socket import ServiceSubtaskRecordSocket
        from .singlepoint.record_socket import SinglepointRecordSocket
        from .optimization.record_socket import OptimizationRecordSocket
        from .torsiondrive.record_socket import TorsiondriveRecordSocket
        from .gridoptimization.record_socket import GridoptimizationRecordSocket
        from .reaction.record_socket import ReactionRecordSocket
        from .manybody.record_socket import ManybodyRecordSocket
        from .neb.record_socket import NEBRecordSocket

        self.service_subtask = ServiceSubtaskRecordSocket(root_socket)
        self.singlepoint = SinglepointRecordSocket(root_socket)
        self.optimization = OptimizationRecordSocket(root_socket)
        self.torsiondrive = TorsiondriveRecordSocket(root_socket)
        self.gridoptimization = GridoptimizationRecordSocket(root_socket)
        self.reaction = ReactionRecordSocket(root_socket)
        self.manybody = ManybodyRecordSocket(root_socket)
        self.neb = NEBRecordSocket(root_socket)

        self._handler_map: Dict[str, BaseRecordSocket] = {
            "servicesubtask": self.service_subtask,
            "singlepoint": self.singlepoint,
            "optimization": self.optimization,
            "torsiondrive": self.torsiondrive,
            "gridoptimization": self.gridoptimization,
            "reaction": self.reaction,
            "manybody": self.manybody,
            "neb": self.neb,
        }

        self._handler_map_by_schema: Dict[str, BaseRecordSocket] = {
            "qcschema_output": self.singlepoint,
            "qcschema_optimization_output": self.optimization,
        }

        self._handler_map_by_input: Dict[Any, BaseRecordSocket] = {
            self.singlepoint.record_input_type: self.singlepoint,
            self.optimization.record_input_type: self.optimization,
            self.torsiondrive.record_input_type: self.torsiondrive,
            self.gridoptimization.record_input_type: self.gridoptimization,
            self.reaction.record_input_type: self.reaction,
            self.manybody.record_input_type: self.manybody,
            self.neb.record_input_type: self.neb,
        }

    def get_socket(self, record_type: str) -> BaseRecordSocket:
        """
        Get the socket for a specific kind of record type
        """

        handler = self._handler_map.get(record_type, None)
        if handler is None:
            raise MissingDataError(f"Cannot find handler for type {record_type}")
        return handler

    def get_dependency_ids(self, session: Session, record_ids: Iterable[int]) -> List[int]:
        """
        Get record ids of dependencies of the given records

        This only makes sense for records that are services. These services have
        records that are dependencies of that service.

        Note that this function takes in and returns record ids, not service queue ids.
        """

        # List may contain duplicates. So be tolerant of that!
        stmt = select(ServiceDependencyORM.record_id)
        stmt = stmt.join(ServiceQueueORM, ServiceQueueORM.id == ServiceDependencyORM.service_id)
        stmt = stmt.where(ServiceQueueORM.record_id.in_(record_ids))
        return session.execute(stmt).scalars().all()

    def get_children_ids(self, session: Session, record_ids: Iterable[int]) -> List[int]:
        """
        Recursively obtain the IDs of all children records of the given records
        """

        # Do in batches to prevent really huge queries
        all_children: List[int] = []
        for id_batch in chunk_iterable(record_ids, 100):
            stmt = select(RecordChildrenView.c.child_id).where(RecordChildrenView.c.parent_id.in_(id_batch))
            children_ids = session.execute(stmt).scalars().all()
            all_children.extend(children_ids)

        return list(set(all_children))

    def get_short_descriptions(
        self, record_ids: Iterable[int], *, session: Optional[Session] = None
    ) -> List[Dict[str, Any]]:
        """
        Obtain short descriptions of records
        """

        stmt = select(BaseRecordORM)
        stmt = stmt.options(defer(BaseRecordORM.properties))  # no need to load these, and they can be big
        stmt = stmt.where(BaseRecordORM.id.in_(record_ids))

        with self.root_socket.optional_session(session, True) as session:
            record_orms = session.execute(stmt).scalars().all()

            found_ids = {r.id for r in record_orms}
            if found_ids != set(record_ids):
                raise MissingDataError("Cannot find all record ids")

            desc_map = {}
            for r in record_orms:
                desc_map[r.id] = {
                    "id": r.id,
                    "record_type": r.record_type,
                    "status": r.status,
                    "created_on": r.created_on,
                    "modified_on": r.modified_on,
                    "description": r.short_description,
                }

            return [desc_map[i] for i in record_ids]

    def get_parent_ids(self, session: Session, record_ids: Iterable[int]) -> List[int]:
        """
        Recursively obtain the IDs of all parent records of the given records
        """

        # Do in batches to prevent really huge queries
        all_parents: List[int] = []

        for id_batch in chunk_iterable(record_ids, 100):
            stmt = select(RecordChildrenView.c.parent_id).where(RecordChildrenView.c.child_id.in_(id_batch))
            parent_ids = session.execute(stmt).scalars().all()
            all_parents.extend(parent_ids)

        return list(set(all_parents))

    def get_relative_ids(self, session: Session, record_ids: Iterable[int]) -> List[int]:
        """
        Recursively obtain the IDs of all parent and children records of the given records
        """

        all_relatives = set()

        records_to_search = set(record_ids)
        while records_to_search:
            direct_relatives = set()

            # do in batches to prevent really huge queries
            for id_batch in chunk_iterable(records_to_search, 100):
                stmt = select(RecordChildrenView.c.parent_id).where(RecordChildrenView.c.child_id.in_(id_batch))
                parent_ids = session.execute(stmt).scalars().unique().all()

                stmt = select(RecordChildrenView.c.child_id).where(RecordChildrenView.c.parent_id.in_(id_batch))
                child_ids = session.execute(stmt).scalars().unique().all()

                direct_relatives.update(parent_ids)
                direct_relatives.update(child_ids)

            # Search through the new direct relatives, but not any we have done already
            records_to_search = direct_relatives - all_relatives
            all_relatives.update(direct_relatives)

        return list(all_relatives)

    def query_base(
        self,
        stmt: Select,
        orm_type: Type[BaseRecordORM],
        query_data: RecordQueryFilters,
        *,
        session: Optional[Session] = None,
    ) -> List[int]:
        """
        Core query functionality of all records

        Each record type will first create an SQLAlchemy Select stmt with all the record-specific
        fields and options. Then it will pass that stmt to this function, which will add queries/filters for all the
        common fields (ids, manager_names, created/modified, etc) that are passed in through
        `query_data`

        Parameters
        ----------
        stmt
            An SQLAlchemy select statement with record-specific filters
        orm_type
            The type of ORM we are querying for
        query_data
            Common record filters to add to the query
        session
            An existing SQLAlchemy session to use. If None, one will be created. If an existing session
            is used, it will be flushed (but not committed) before returning from this function.

        Returns
        -------
        :
            A list of record ids that were found in the database.
        """

        and_query = []
        if query_data.record_id is not None:
            and_query.append(orm_type.id.in_(query_data.record_id))
        if query_data.record_type is not None:
            and_query.append(orm_type.record_type.in_(query_data.record_type))
        if query_data.manager_name is not None:
            and_query.append(orm_type.manager_name.in_(query_data.manager_name))
        if query_data.status is not None:
            and_query.append(orm_type.status.in_(query_data.status))
        if query_data.created_before is not None:
            and_query.append(orm_type.created_on <= query_data.created_before)
        if query_data.created_after is not None:
            and_query.append(orm_type.created_on >= query_data.created_after)
        if query_data.modified_before is not None:
            and_query.append(orm_type.modified_on <= query_data.modified_before)
        if query_data.modified_after is not None:
            and_query.append(orm_type.modified_on >= query_data.modified_after)

        if query_data.creator_user is not None:
            stmt = stmt.join(UserIDMapSubquery)

            int_ids = {x for x in query_data.creator_user if isinstance(x, int) or x.isdecimal()}
            str_names = set(query_data.creator_user) - int_ids

            and_query.append(or_(UserIDMapSubquery.username.in_(str_names), UserIDMapSubquery.id.in_(int_ids)))

        if query_data.parent_id is not None:
            # We alias the cte because we might join on it twice
            parent_cte = aliased(RecordDirectChildrenView)

            stmt = stmt.join(parent_cte, parent_cte.c.child_id == orm_type.id)
            stmt = stmt.where(parent_cte.c.parent_id.in_(query_data.parent_id))

        if query_data.child_id is not None:
            # We alias the cte because we might join on it twice
            child_cte = aliased(RecordDirectChildrenView)

            stmt = stmt.join(child_cte, child_cte.c.parent_id == orm_type.id)
            stmt = stmt.where(child_cte.c.child_id.in_(query_data.child_id))

        if query_data.dataset_id is not None:
            stmt = stmt.join(DatasetDirectRecordsView, DatasetDirectRecordsView.c.record_id == orm_type.id)
            stmt = stmt.where(DatasetDirectRecordsView.c.dataset_id.in_(query_data.dataset_id))

        with self.root_socket.optional_session(session, True) as session:
            stmt = stmt.where(*and_query)

            if query_data.cursor is not None:
                stmt = stmt.where(orm_type.id < query_data.cursor)

            stmt = stmt.order_by(orm_type.id.desc())
            stmt = stmt.limit(query_data.limit)
            stmt = stmt.distinct(orm_type.id)
            record_ids = session.execute(stmt).scalars().all()

        return record_ids

    def query(
        self,
        query_data: RecordQueryFilters,
        *,
        session: Optional[Session] = None,
    ) -> List[int]:
        """
        Query records of all types based on common fields

        Parameters
        ----------
        query_data
            Fields/filters to query for
        session
            An existing SQLAlchemy session to use. If None, one will be created. If an existing session
            is used, it will be flushed (but not committed) before returning from this function.

        Returns
        -------
        :
            A list of record ids that were found in the database.
        """

        wp = with_polymorphic(BaseRecordORM, "*")
        stmt = select(wp.id)

        return self.query_base(
            stmt,
            orm_type=wp,
            query_data=query_data,
            session=session,
        )

    def get_base(
        self,
        orm_type: Type[BaseRecordORM],
        record_ids: Sequence[int],
        include: Optional[Sequence[str]] = None,
        exclude: Optional[Sequence[str]] = None,
        missing_ok: bool = False,
        additional_options: Optional[List[Any]] = None,
        *,
        session: Optional[Session] = None,
    ):
        options = []
        if include is not None:
            include_history = is_included("compute_history", include, exclude, False)
            include_outputs = is_included("outputs", include, exclude, False)

            if include_history or include_outputs:
                options.append(selectinload(orm_type.compute_history))
            if include_outputs:
                options.append(
                    selectinload(orm_type.compute_history, RecordComputeHistoryORM.outputs).undefer(OutputStoreORM.data)
                )
            if is_included("task", include, exclude, False):
                options.append(joinedload(orm_type.task))
            if is_included("service", include, exclude, False):
                options.append(joinedload(orm_type.service))
            if is_included("comments", include, exclude, False):
                options.append(selectinload(orm_type.comments))
            if is_included("native_files", include, exclude, False):
                options.append(selectinload(orm_type.native_files).options(undefer(NativeFileORM.data)))

        if additional_options:
            options.extend(additional_options)

        with self.root_socket.optional_session(session, True) as session:
            return get_general(session, orm_type, orm_type.id, record_ids, include, exclude, missing_ok, options)

    def get(
        self,
        record_ids: Sequence[int],
        include: Optional[Sequence[str]] = None,
        exclude: Optional[Sequence[str]] = None,
        missing_ok: bool = False,
        *,
        session: Optional[Session] = None,
    ) -> List[Optional[Dict[str, Any]]]:
        """
        Obtain records of all types with specified IDs

        Parameters
        ----------
        record_ids
            A list or other sequence of record IDs
        include
            Which fields of the result to return. Default is to return all fields.
        exclude
            Remove these fields from the return. Default is to return all fields.
        missing_ok
           If set to True, then missing results will be tolerated, and the returned list of
           records will contain None for the corresponding IDs that were not found.
        session
            An existing SQLAlchemy session to use. If None, one will be created. If an existing session
            is used, it will be flushed (but not committed) before returning from this function.

        Returns
        -------
        :
            Records (as dictionaries) in the same order as the given ids.
            If missing_ok is True, then this list will contain None where the record was missing.
        """

        # If all columns are included, then we can load
        # the data from derived classes as well.
        if (include is None or "*" in include or "**" in include) and not exclude:
            wp = with_polymorphic(BaseRecordORM, "*")
        else:
            wp = BaseRecordORM

        return self.get_base(wp, record_ids, include, exclude, missing_ok, session=session)

    def initialize_service(self, session: Session, service_orm: ServiceQueueORM) -> None:
        """
        Initialize a service

        A service is initialized when it moves from a waiting state to a running state. After it is
        initialized, iterate_service may be called.
        """

        if not service_orm.record.is_service:
            raise RuntimeError("Cannot initialize a record that is not a service")

        record_type = service_orm.record.record_type
        service_socket = self._handler_map[record_type]

        if not service_socket.available():
            self._logger.warning(
                f"Cannot initialize service id={service_orm.id} - socket is not available."
                "Are the correct packages installed?"
            )
        else:
            service_socket.initialize_service(session, service_orm)

    def iterate_service(self, session: Session, service_orm: ServiceQueueORM) -> bool:
        """
        Iterate a service with successfully-completed dependencies

        The server will check to see if all dependencies are successfully completed. If they are, this function
        is called.

        Cases where the service has an errored dependency are handled in :meth:`RecordSocket.update_failed_service`.
        """

        if not service_orm.record.is_service:
            raise RuntimeError("Cannot iterate a record that is not a service")

        record_type = service_orm.record.record_type
        service_socket = self._handler_map[record_type]

        if not service_socket.available():
            self._logger.warning(
                f"Cannot iterate service id={service_orm.id} - socket is not available."
                "Are the correct packages installed?"
            )
            return False

        # Actually iterate the service
        finished = service_socket.iterate_service(session, service_orm)

        if not finished:
            # Update the sort_date on any tasks that were created
            # The sort date should almost always be the created_on of the parent service
            # This way, earlier services will have their tasks run first, rather than having to wait
            # for all other services' tasks to finish
            # But, it could be the original submission date of the record
            parent_created_on = service_orm.record.created_on

            stmt = """
                WITH tq_cte AS (
                    SELECT tq.id FROM task_queue tq
                    INNER JOIN service_dependency sd ON tq.record_id = sd.record_id
                    WHERE sd.service_id = :service_id
                    AND tq.available = true
                    AND tq.sort_date > :parent_created_on
                    FOR UPDATE OF tq SKIP LOCKED
                )
                UPDATE task_queue AS tq
                SET sort_date = :parent_created_on
                FROM tq_cte
                WHERE tq_cte.id = tq.id
            """

            session.execute(text(stmt), {"parent_created_on": parent_created_on, "service_id": service_orm.id})

        return finished

    def update_failed_service(self, session, record_orm: BaseRecordORM, error_info: Dict[str, Any]):
        """
        Handle a service with errored dependencies
        """
        record_orm.status = RecordStatusEnum.error

        error_orm = create_output_orm(OutputTypeEnum.error, error_info)
        record_orm.compute_history[-1].status = RecordStatusEnum.error

        upsert_output(session, record_orm, error_orm)
        record_orm.compute_history[-1].modified_on = now_at_utc()

        record_orm.status = RecordStatusEnum.error
        record_orm.modified_on = now_at_utc()

    ######################
    # Non-service records
    ######################
    def update_failed_task(self, session: Session, record_id: int, failed_result: FailedOperation, manager_name: str):
        """
        Update a record ORM whose computation has failed, and mark it as errored

        Parameters
        ----------
        session
            An existing SQLAlchemy session
        record_id
            ID of the record to update and mark as completed
        failed_result
            The (failed) result of the computation
        manager_name
            The manager that produced the result
        """

        if not isinstance(failed_result, FailedOperation):
            raise RuntimeError("Developer error - this function only handles FailedOperation results")

        # Handle outputs, which are a bit different in FailedOperation
        # Error is special in a FailedOperation
        error = failed_result.error
        if error is None:
            error = _default_error

        error_out_orm = create_output_orm(OutputTypeEnum.error, error.dict())
        all_outputs = {OutputTypeEnum.error: error_out_orm}

        # Get the rest of the outputs
        # This is stored in "input_data" (I know...)
        # "input_data" can be anything. So ignore if it isn't a dict
        if isinstance(failed_result.input_data, dict):
            stdout = failed_result.input_data.get("stdout", None)
            stderr = failed_result.input_data.get("stderr", None)

            if stdout is not None:
                stdout_orm = create_output_orm(OutputTypeEnum.stdout, stdout)
                all_outputs[OutputTypeEnum.stdout] = stdout_orm
            if stderr is not None:
                stderr_orm = create_output_orm(OutputTypeEnum.stderr, stderr)
                all_outputs[OutputTypeEnum.stderr] = stderr_orm

        # Build the history orm
        history_orm = RecordComputeHistoryORM()
        history_orm.status = RecordStatusEnum.error
        history_orm.manager_name = manager_name
        history_orm.modified_on = now_at_utc()
        history_orm.outputs = all_outputs
        history_orm.record_id = record_id
        session.add(history_orm)

        # What updates we have for this record
        record_updates = {
            "status": RecordStatusEnum.error,
            "modified_on": history_orm.modified_on,
            "manager_name": manager_name,
        }

        # Actually update the record
        stmt = update(BaseRecordORM).where(BaseRecordORM.id == record_id).values(record_updates)
        session.execute(stmt)

    def add_from_input(
        self,
        record_input: AllInputTypes,
        compute_tag: str,
        compute_priority: PriorityEnum,
        creator_user: Optional[Union[int, str]],
        find_existing: bool,
        *,
        session: Optional[Session] = None,
    ):
        # TODO - placeholder for now. Could support other types of inputs
        return self.add_from_input_qcportal(
            record_input, compute_tag, compute_priority, creator_user, find_existing, session=session
        )

    def update_completed(
        self, session: Session, record_id: int, record_type: str, result: AllResultTypes, manager_name: str
    ):
        return self.update_completed_schema_v1(session, record_id, record_type, result, manager_name)

    def insert_full_record(
        self, session: Session, results: Sequence[AllResultTypes], creator_user: Optional[Union[int, str]]
    ) -> List[int]:

        all_ids: List[Optional[int]] = [None for _ in range(len(results))]

        schema_v1_results: List[Tuple[int, AllSchemaV1ResultTypes]] = []
        qcp_results: List[Tuple[int, AllQCPortalRecordTypes]] = []

        for idx, result in enumerate(results):
            is_qcptype = isinstance(result, AllQCPortalRecordTypes)
            is_schemav1type = isinstance(result, AllSchemaV1ResultTypes)

            if isinstance(result, FailedOperation) or (is_schemav1type and not result.success):
                raise UserReportableError("Cannot insert a full, failed operation")
            elif is_schemav1type:
                schema_v1_results.append((idx, result))
            elif is_qcptype:
                qcp_results.append((idx, result))
            else:
                raise UserReportableError("Cannot insert a completed, unrecognized result")

        creator_user_id = self.root_socket.users.get_optional_user_id(creator_user, session=session)

        if schema_v1_results:
            schema_tmp = [x[1] for x in schema_v1_results]
            schema_ids = self.insert_full_schema_v1(session, schema_tmp, creator_user_id)
            for (schema_idx, _), schema_id in zip(schema_v1_results, schema_ids):
                all_ids[schema_idx] = schema_id

        if qcp_results:
            qcp_tmp = [x[1] for x in qcp_results]
            qcp_ids = self.insert_full_qcportal_records(session, qcp_tmp, creator_user_id)
            for (record_idx, _), record_id in zip(qcp_results, qcp_ids):
                all_ids[record_idx] = record_id

        if None in all_ids:
            raise RuntimeError("Developer error - Missing record ID in insert_full_record")

        return all_ids

    ######################################################
    # Handlers for different result types (schema, etc)
    ######################################################
    def add_from_input_qcportal(
        self,
        record_input: AllQCPortalInputTypes,
        compute_tag: str,
        compute_priority: PriorityEnum,
        creator_user: Optional[Union[int, str]],
        find_existing: bool,
        *,
        session: Optional[Session] = None,
    ):
        record_socket = self._handler_map_by_input[type(record_input)]
        return record_socket.add_from_input(
            record_input, compute_tag, compute_priority, creator_user, find_existing, session=session
        )

    def update_completed_schema_v1(
        self, session: Session, record_id: int, record_type: str, result: AllSchemaV1ResultTypes, manager_name: str
    ):
        """
        Update a record ORM based on the results of a successfully-completed computation

        Parameters
        ----------
        session
            An existing SQLAlchemy session
        record_id
            ID of the record to update and mark as completed
        record_type
            Type of record to update
        result
            The result of the computation. This should be a successful result
        manager_name
            The manager that produced the result
        """

        # Do these before calling the record-specific handler
        # (these may pull stuff out of extras)
        history_orm = compute_history_orm_from_schema_v1(result)
        history_orm.record_id = record_id
        history_orm.manager_name = manager_name
        session.add(history_orm)

        native_files_orms = native_files_orms_from_schema_v1(result)
        for v in native_files_orms.values():
            v.record_id = record_id
            session.add(v)

        # Now update fields specific to each record
        record_socket = self._handler_map[record_type]
        record_socket.update_completed_schema_v1(session, record_id, result)

        # Now extras and properties
        extras, properties = build_extras_properties(result)

        # What updates we have for this record
        record_updates = {
            "extras": extras,
            "properties": properties,
            "status": RecordStatusEnum.complete,
            "manager_name": manager_name,
            "modified_on": history_orm.modified_on,
        }

        # Actually update the record
        stmt = update(BaseRecordORM).where(BaseRecordORM.id == record_id).values(record_updates)
        session.execute(stmt)

        # Delete the task from the task queue since it is completed
        stmt = delete(TaskQueueORM).where(TaskQueueORM.record_id == record_id)
        session.execute(stmt)

    def insert_full_qcportal_records(
        self, session: Session, records: Sequence[AllQCPortalRecordTypes], creator_user_id: Optional[int]
    ) -> List[int]:
        """
        Insert records into the database from a QCPortal record

        This will always create new ORMs from scratch, and not update any existing records.
        """

        if len(records) == 0:
            return []

        # Map of schema name stored in the records to (index, record)
        type_map: Dict[str, List[Tuple[int, AllQCPortalRecordTypes]]] = defaultdict(list)

        for idx, record in enumerate(records):
            type_map[record.record_type].append((idx, record))

        # First element in the tuple is the index in the original "records" list
        to_add: List[Tuple[int, BaseRecordORM]] = []

        for record_type, idx_records in type_map.items():
            # idx_records is a list of tuple (idx, record). idx is the index in the
            # original records (function parameter)
            type_records = [r for _, r in idx_records]
            history_orms = [compute_history_orms_from_qcportal_record(r) for r in type_records]
            native_files_orms = [native_files_orms_from_qcportal_record(r) for r in type_records]

            # Now the record-specific stuff
            handler = self._handler_map[record_type]
            record_orms = handler.insert_full_qcportal_records_v1(session, type_records, creator_user_id)

            for record_orm, history_orm, native_files_orm, (idx, type_record) in zip(
                record_orms, history_orms, native_files_orms, idx_records
            ):
                record_orm.creator_user_id = creator_user_id
                record_orm.is_service = type_record.is_service

                # Now back to the common modifications
                record_orm.compute_history = history_orm
                record_orm.native_files = native_files_orm
                record_orm.status = type_record.status

                if type_record.comments_:
                    record_orm.comments = [
                        RecordCommentORM(
                            timestamp=c.timestamp,
                            comment=c.comment,
                        )
                        for c in type_record.comments_
                    ]

                record_orm.created_on = type_record.created_on
                record_orm.modified_on = type_record.modified_on
                record_orm.manager_name = type_record.manager_name

                record_orm.extras = type_record.extras
                record_orm.properties = type_record.properties

                to_add.append((idx, record_orm))

        to_add.sort()

        # Check that everything is in the original order
        assert tuple(idx for idx, _ in to_add) == tuple(range(len(to_add)))
        to_add_orm = [o for _, o in to_add]

        session.add_all(to_add_orm)
        session.flush()

        return [o.id for o in to_add_orm]

    def insert_full_schema_v1(
        self, session: Session, results: Sequence[AllSchemaV1ResultTypes], creator_user_id: Optional[int]
    ) -> List[int]:
        """
        Insert records into the database from a QCSchema result

        This will always create new ORMs from scratch, and not update any existing records.
        """

        if len(results) == 0:
            return []

        # Map of schema name stored in the result to (index, result)
        type_map: Dict[str, List[Tuple[int, AllSchemaV1ResultTypes]]] = defaultdict(list)

        for idx, result in enumerate(results):
            if isinstance(result, FailedOperation) or not result.success:
                raise UserReportableError("Cannot insert a completed, failed operation")

            type_map[result.schema_name].append((idx, result))

        # Key is the index in the original "results" dict
        to_add: List[Tuple[int, BaseRecordORM]] = []

        for schema_name, idx_results in type_map.items():
            # Get the outputs & status, storing in the history orm.
            # Do this before calling the individual record handlers since it modifies extras
            # (looking for compressed outputs and compressed native files)

            # idx_results is a list of tuple (idx, result). idx is the index in the
            # original results (function parameter)
            type_results = [r for _, r in idx_results]
            history_orms = [compute_history_orm_from_schema_v1(r) for r in type_results]
            native_files_orms = [native_files_orms_from_schema_v1(r) for r in type_results]

            # Now the record-specific stuff
            handler = self._handler_map_by_schema[schema_name]
            record_orms = handler.insert_full_schema_v1(session, type_results, creator_user_id)

            for record_orm, history_orm, native_files_orm, (idx, type_result) in zip(
                record_orms, history_orms, native_files_orms, idx_results
            ):
                record_orm.creator_user_id = creator_user_id
                record_orm.is_service = False

                # Now extras and properties
                extras, properties = build_extras_properties(type_result)
                record_orm.extras = extras
                record_orm.properties = properties

                # Now back to the common modifications
                record_orm.compute_history.append(history_orm)
                record_orm.native_files = native_files_orm
                record_orm.status = history_orm.status
                record_orm.modified_on = history_orm.modified_on

                to_add.append((idx, record_orm))

        to_add.sort()

        # Check that everything is in the original order
        assert tuple(idx for idx, _ in to_add) == tuple(range(len(to_add)))
        to_add_orm = [o for _, o in to_add]

        session.add_all(to_add_orm)
        session.flush()

        return [o.id for o in to_add_orm]

    ########################################
    # Various record manipulation functions
    ########################################
    def add_comment(
        self, record_ids: Sequence[int], user_id: Optional[int], comment: str, *, session: Optional[Session] = None
    ) -> UpdateMetadata:
        """
        Adds a comment to records

        Comments are free-form strings attached to a record.

        Parameters
        ----------
        record_ids
            Records to add the comment to
        user_id
            The ID of the user that is adding the comment
        comment
            The comment to be attached to the record
        session
            An existing SQLAlchemy session to use. If None, one will be created. If an existing session
            is used, it will be flushed before returning from this function.

        Returns
        -------
        :
            Metadata about the modification/update.
        """

        with self.root_socket.optional_session(session) as session:
            # find only existing records
            stmt = select(BaseRecordORM.id).where(BaseRecordORM.id.in_(record_ids))
            stmt = stmt.with_for_update()
            existing_ids = session.execute(stmt).scalars().all()

            for rid in existing_ids:
                comment_orm = RecordCommentORM(
                    record_id=rid,
                    user_id=user_id,
                    comment=comment,
                )
                session.add(comment_orm)

            updated_idx = [idx for idx, rid in enumerate(record_ids) if rid in existing_ids]
            missing_idx = [idx for idx, rid in enumerate(record_ids) if rid not in existing_ids]

            return UpdateMetadata(
                updated_idx=updated_idx,
                errors=[(idx, "Record does not exist") for idx in missing_idx],
            )

    def reset_assigned(
        self,
        manager_name: Optional[Iterable[str]] = None,
        *,
        session: Optional[Session] = None,
    ) -> List[int]:
        """
        Reset the status of records assigned to given managers to waiting

        Parameters
        ----------
        manager_name
            Reset the status of records belonging to these managers
        session
            An existing SQLAlchemy session to use. If None, one will be created. If an existing session
            is used, it will be flushed (but not committed) before returning from this function.

        Returns
        -------
        :
            Ids of the tasks that were modified
        """

        if not manager_name:
            return []

        with self.root_socket.optional_session(session) as session:
            stmt = select(BaseRecordORM).options(joinedload(BaseRecordORM.task, innerjoin=True))
            stmt = stmt.where(BaseRecordORM.is_service.is_(False))
            stmt = stmt.where(BaseRecordORM.manager_name.in_(manager_name))
            stmt = stmt.where(BaseRecordORM.status == RecordStatusEnum.running)
            stmt = stmt.with_for_update()

            record_orms = session.execute(stmt).scalars().all()

            for r in record_orms:
                r.status = RecordStatusEnum.waiting
                r.modified_on = now_at_utc()
                r.manager_name = None
                r.task.available = True

            return [r.id for r in record_orms]

    def _revert_common(
        self,
        record_ids: Sequence[int],
        applicable_status: Iterable[RecordStatusEnum],
        *,
        session: Optional[Session] = None,
    ) -> UpdateMetadata:
        """
        Internal function for resetting, uncancelling, undeleting, or uninvalidation

        This will also re-create tasks & services as necessary

        Parameters
        ----------
        record_ids
            Reset the status of these record ids
        session
            An existing SQLAlchemy session to use. If None, one will be created. If an existing session
            is used, it will be flushed (but not committed) before returning from this function.

        Returns
        -------
        :
            Metadata about what was updated
        """

        if not record_ids:
            return UpdateMetadata()

        all_ids = set(record_ids)

        with self.root_socket.optional_session(session) as session:
            # We always apply these operations to children, but never to parents
            children_ids = self.get_children_ids(session, record_ids)
            all_ids.update(children_ids)

            updated_ids = []
            for id_batch in chunk_iterable(all_ids, 100):
                # Select records with a resettable status
                # All are resettable except complete and running
                # Can't do inner join because task may not exist
                stmt = select(BaseRecordORM).options(
                    joinedload(BaseRecordORM.task),
                    load_only(BaseRecordORM.id, BaseRecordORM.status, BaseRecordORM.manager_name),
                )
                stmt = stmt.options(selectinload(BaseRecordORM.info_backup))
                stmt = stmt.where(BaseRecordORM.status.in_(applicable_status))
                stmt = stmt.where(BaseRecordORM.id.in_(id_batch))
                stmt = stmt.with_for_update(of=[BaseRecordORM])
                record_data = session.execute(stmt).scalars().all()

                for r_orm in record_data:
                    # If we have the old backup info, then use that
                    if (
                        r_orm.status in [RecordStatusEnum.deleted, RecordStatusEnum.cancelled, RecordStatusEnum.invalid]
                        and r_orm.info_backup
                    ):
                        last_info = r_orm.info_backup.pop()  # Remove the last entry
                        r_orm.status = last_info.old_status

                        if r_orm.status in [RecordStatusEnum.waiting, RecordStatusEnum.error]:
                            if r_orm.task:
                                self._logger.warning(
                                    f"Record {r_orm.id} has a task and also an entry in the backup table!"
                                )
                                session.delete(r_orm.task)

                            # we leave service queue entries alone
                            if not r_orm.is_service:
                                BaseRecordSocket.create_task(
                                    r_orm, last_info.old_compute_tag, last_info.old_compute_priority
                                )

                    elif r_orm.status in [RecordStatusEnum.running, RecordStatusEnum.error] and not r_orm.info_backup:
                        if not r_orm.is_service and r_orm.task is None:
                            raise RuntimeError(f"resetting a record with status {r_orm.status} with no task")
                        if r_orm.is_service and r_orm.service is None:
                            raise RuntimeError(f"resetting a record with status {r_orm.status} with no service")

                        # Move the record back to "waiting" for a manager/service periodics to pick it up
                        r_orm.status = RecordStatusEnum.waiting
                        r_orm.manager_name = None

                        if not r_orm.is_service:
                            r_orm.task.available = True

                    else:
                        if r_orm.info_backup:
                            raise RuntimeError(f"resetting record with status {r_orm.status} with backup info present")
                        else:
                            raise RuntimeError(
                                f"resetting record with status {r_orm.status} without backup info present"
                            )

                    r_orm.modified_on = now_at_utc()

                    # Sanity check
                    if not r_orm.is_service:
                        if r_orm.status == RecordStatusEnum.waiting:
                            assert r_orm.task.available == True
                        else:
                            assert r_orm.task is None or r_orm.task.available == False

                updated_ids.extend([r.id for r in record_data])

            # put in order of the input parameter
            error_ids = set(record_ids) - set(updated_ids)
            updated_idx = [idx for idx, rid in enumerate(record_ids) if rid in updated_ids]
            error_idx = [idx for idx, rid in enumerate(record_ids) if rid in error_ids]
            errors = [(idx, "Record is missing or cannot be reset") for idx in error_idx]
            n_children_updated = len(updated_ids) - len(updated_idx)

            return UpdateMetadata(updated_idx=updated_idx, errors=errors, n_children_updated=n_children_updated)

    def _cancel_common(
        self,
        record_ids: Sequence[int],
        applicable_status: Iterable[RecordStatusEnum],
        new_status: RecordStatusEnum,
        propagate_to_children: bool,
        *,
        session: Optional[Session] = None,
    ) -> UpdateMetadata:
        """
        Internal function for cancelling, deleting, or invalidation

        Cancelling, deleting, and invalidation basically the same, with the only difference
        being that they apply to different statuses.
        The connotation of these operations is different, but internally they behave the same.

        Parameters
        ----------
        record_ids
            Reset the status of these record ids
        new_status
            What the new status of the record should be
        propagate_to_children
            Apply the operation to children as well
        session
            An existing SQLAlchemy session to use. If None, one will be created. If an existing session
            is used, it will be flushed (but not committed) before returning from this function.

        Returns
        -------
        :
            Metadata about what was updated
        """

        if len(record_ids) == 0:
            return UpdateMetadata()

        all_ids = set(record_ids)

        with self.root_socket.optional_session(session) as session:
            if propagate_to_children:
                # We always propagate to parents. So recursively find all related records
                relative_ids = self.get_relative_ids(session, record_ids)
                all_ids.update(relative_ids)
            else:
                # Always propagate these changes to parents
                parent_ids = self.get_parent_ids(session, record_ids)
                all_ids.update(parent_ids)

            updated_ids = []
            for id_batch in chunk_iterable(all_ids, 100):
                stmt = select(BaseRecordORM).options(
                    joinedload(BaseRecordORM.task),
                    load_only(BaseRecordORM.id, BaseRecordORM.status, BaseRecordORM.manager_name),
                )
                stmt = stmt.where(BaseRecordORM.status.in_(applicable_status))
                stmt = stmt.where(BaseRecordORM.id.in_(id_batch))
                stmt = stmt.with_for_update(of=[BaseRecordORM])
                record_orms = session.execute(stmt).scalars().all()

                for r in record_orms:
                    # If running, delete the manager. Resetting later will move it to waiting
                    if r.status == RecordStatusEnum.running:
                        r.status = RecordStatusEnum.waiting
                        r.manager_name = None

                    old_tag = None
                    old_priority = None
                    if r.task is not None:
                        old_tag = r.task.compute_tag
                        old_priority = r.task.compute_priority
                        session.delete(r.task)

                    # If this is a service, we leave the
                    # the entry in the service queue since it contains
                    # the current service state info

                    # Store the old info in the backup table
                    backup_info = RecordInfoBackupORM(
                        record_id=r.id,
                        old_status=r.status,
                        old_compute_tag=old_tag,
                        old_compute_priority=old_priority,
                        modified_on=now_at_utc(),
                    )
                    session.add(backup_info)

                    r.modified_on = now_at_utc()
                    r.status = new_status

                updated_ids.extend([r.id for r in record_orms])

            # put in order of the input parameter
            error_ids = set(record_ids) - set(updated_ids)
            updated_idx = [idx for idx, rid in enumerate(record_ids) if rid in updated_ids]
            error_idx = [idx for idx, rid in enumerate(record_ids) if rid in error_ids]
            errors = [(idx, "Record is missing or cannot be cancelled/deleted/invalidated") for idx in error_idx]
            n_children_updated = len(updated_ids) - len(updated_idx)

            return UpdateMetadata(updated_idx=updated_idx, errors=errors, n_children_updated=n_children_updated)

    def reset(self, record_ids: Sequence[int], *, session: Optional[Session] = None):
        """
        Resets errored records to be waiting again
        """

        return self._revert_common(record_ids, applicable_status=[RecordStatusEnum.error], session=session)

    def reset_running(self, record_ids: Sequence[int], *, session: Optional[Session] = None):
        """
        Resets running records to be waiting again
        """

        return self._revert_common(record_ids, applicable_status=[RecordStatusEnum.running], session=session)

    def delete(
        self,
        record_ids: Sequence[int],
        soft_delete: bool = True,
        delete_children: bool = True,
        *,
        session: Optional[Session] = None,
    ) -> DeleteMetadata:
        """
        Delete records from the database

        If soft_delete is True, then the record is just marked as deleted and actually deletion may
        happen later. Soft delete can be undone with undelete

        A deleted record will not be picked up by any manager.

        Parameters
        ----------
        record_ids
            Reset the status of these record ids
        soft_delete
            Don't actually delete the record, just mark it for later deletion
        delete_children
            If True, attempt to delete child records as well
        session
            An existing SQLAlchemy session to use. If None, one will be created. If an existing session
            is used, it will be flushed (but not committed) before returning from this function.

        Returns
        -------
        :
            Metadata about what was deleted (or marked as deleted)
        """

        if len(record_ids) == 0:
            return DeleteMetadata()

        if soft_delete:
            # anything can be deleted except something already deleted
            meta = self._cancel_common(
                record_ids,
                set(RecordStatusEnum) - {RecordStatusEnum.deleted},
                RecordStatusEnum.deleted,
                propagate_to_children=delete_children,
                session=session,
            )

            # convert the update metadata to a deleted metadata
            return DeleteMetadata(
                error_description=meta.error_description,
                errors=meta.errors,
                deleted_idx=meta.updated_idx,
                n_children_deleted=meta.n_children_updated,
            )

        with self.root_socket.optional_session(session) as session:
            children_ids = []

            if delete_children:
                children_ids = self.get_children_ids(session, record_ids)

            del_id_1 = [(x,) for x in record_ids]
            del_id_2 = [(x,) for x in children_ids]
            meta = delete_general(session, BaseRecordORM, BaseRecordORM.id, del_id_1)
            ch_meta = delete_general(session, BaseRecordORM, BaseRecordORM.id, del_id_2)

            meta_dict = meta.dict()
            meta_dict["n_children_deleted"] = ch_meta.n_deleted
            return DeleteMetadata(**meta_dict)

    def cancel(
        self,
        record_ids: Sequence[int],
        cancel_children: bool = True,
        *,
        session: Optional[Session] = None,
    ) -> UpdateMetadata:
        """
        Marks records as cancelled

        A cancelled record will not be picked up by a manager.

        Parameters
        ----------
        record_ids
            Reset the status of these record ids
        cancel_children
            Cancel all children as well
        session
            An existing SQLAlchemy session to use. If None, one will be created. If an existing session
            is used, it will be flushed (but not committed) before returning from this function.

        Returns
        -------
        :
            Metadata about what was deleted
        """

        return self._cancel_common(
            record_ids,
            {RecordStatusEnum.waiting, RecordStatusEnum.running, RecordStatusEnum.error},
            RecordStatusEnum.cancelled,
            propagate_to_children=cancel_children,
            session=session,
        )

    def invalidate(
        self,
        record_ids: Sequence[int],
        *,
        session: Optional[Session] = None,
    ) -> UpdateMetadata:
        """
        Marks records as invalid

        An invalid record is one that supposedly successfully completed. However, after review,
        is not correct.

        Parameters
        ----------
        record_ids
            Reset the status of these record ids
        session
            An existing SQLAlchemy session to use. If None, one will be created. If an existing session
            is used, it will be flushed (but not committed) before returning from this function.

        Returns
        -------
        :
            Metadata about what was deleted
        """

        return self._cancel_common(
            record_ids,
            {RecordStatusEnum.complete},
            RecordStatusEnum.invalid,
            propagate_to_children=False,
            session=session,
        )

    def undelete(
        self,
        record_ids: Sequence[int],
        *,
        session: Optional[Session] = None,
    ) -> UpdateMetadata:
        """
        Undeletes records that were soft deleted

        This will always undelete children whenever possible
        """
        if len(record_ids) == 0:
            return UpdateMetadata()

        with self.root_socket.optional_session(session) as session:
            meta = self._revert_common(record_ids, applicable_status=[RecordStatusEnum.deleted], session=session)

            return UpdateMetadata(
                updated_idx=meta.updated_idx,
                error_description=meta.error_description,
                errors=meta.errors,
                n_children_updated=meta.n_children_updated,
            )

    def uncancel(
        self,
        record_ids: Sequence[int],
        *,
        session: Optional[Session] = None,
    ) -> UpdateMetadata:
        """
        Uncancels records that were previously cancelled

        This will always uncancel children whenever possible
        """
        if len(record_ids) == 0:
            return UpdateMetadata()

        with self.root_socket.optional_session(session) as session:
            return self._revert_common(record_ids, applicable_status=[RecordStatusEnum.cancelled], session=session)

    def uninvalidate(
        self,
        record_ids: Sequence[int],
        *,
        session: Optional[Session] = None,
    ) -> UpdateMetadata:
        """
        Uninvalidates records that were previously uninvalidated

        This will always uninvalidate children whenever possible
        """
        if len(record_ids) == 0:
            return UpdateMetadata()

        with self.root_socket.optional_session(session) as session:
            return self._revert_common(record_ids, applicable_status=[RecordStatusEnum.invalid], session=session)

    def modify(
        self,
        record_ids: Sequence[int],
        new_compute_tag: Optional[str] = None,
        new_compute_priority: Optional[RecordStatusEnum] = None,
        *,
        session: Optional[Session] = None,
    ) -> UpdateMetadata:
        """
        Modifies a record's task's priority and tag

        Records without a corresponding task (completed, etc) will not be modified. Running tasks
        will also not be modified.

        An empty string for new_compute_tag will be treated the same as if you had passed it None

        Parameters
        ----------
        record_ids
            Modify the tasks corresponding to these record ids
        new_compute_tag
            New tag for the task. If None, keep the existing tag
        new_compute_priority
            New priority for the task. If None, then keep the existing priority
        session
            An existing SQLAlchemy session to use. If None, one will be created. If an existing session
            is used, it will be flushed before returning from this function.

        Returns
        -------
        :
            Metadata about what was updated
        """

        if new_compute_tag == "":
            new_compute_tag = None

        # Do we have anything to do?
        if new_compute_tag is None and new_compute_priority is None:
            return UpdateMetadata()

        all_id = set(record_ids)

        with self.root_socket.optional_session(session) as session:
            # Get subtasks for all the records that are services
            subtask_id = self.get_dependency_ids(session, record_ids)
            all_id.update(subtask_id)

            # Do a manual join, not a joined load - we don't want to actually load the base record, just
            # query by status
            stmt = select(TaskQueueORM)
            stmt = stmt.join(TaskQueueORM.record)
            stmt = stmt.where(TaskQueueORM.record_id.in_(all_id))
            stmt = stmt.with_for_update()
            task_orms = session.execute(stmt).scalars().all()

            stmt = select(ServiceQueueORM)
            stmt = stmt.join(ServiceQueueORM.record)
            stmt = stmt.where(ServiceQueueORM.record_id.in_(all_id))
            stmt = stmt.with_for_update()
            svc_orms = session.execute(stmt).scalars().all()

            all_orm = task_orms + svc_orms
            for o in all_orm:
                if new_compute_tag is not None:
                    o.compute_tag = new_compute_tag
                if new_compute_priority is not None:
                    o.compute_priority = new_compute_priority

            # put in order of the input parameter
            # only pay attention to the records requested (ie, not subtasks)
            updated_ids = [t.record_id for t in all_orm]
            error_ids = set(record_ids) - set(updated_ids)
            updated_idx = [idx for idx, rid in enumerate(record_ids) if rid in updated_ids]
            error_idx = [idx for idx, rid in enumerate(record_ids) if rid in error_ids]
            errors = [(idx, "Record is missing or cannot be modified") for idx in error_idx]
            n_children_updated = len(all_orm) - (len(updated_idx) + len(error_idx))

            return UpdateMetadata(updated_idx=updated_idx, errors=errors, n_children_updated=n_children_updated)

    def modify_generic(
        self,
        record_ids: Sequence[int],
        user_id: Optional[int],
        status: Optional[RecordStatusEnum] = None,
        compute_priority: Optional[PriorityEnum] = None,
        compute_tag: Optional[str] = None,
        comment: Optional[str] = None,
        *,
        session: Optional[Session] = None,
    ) -> UpdateMetadata:
        """
        Modify multiple things about a record at once

        This function allows for changing the status, tag, priority, and comment in a single function call

        Parameters
        ----------
        record_ids
            IDs of the records to modify
        user_id
            ID of the user modifying the records
        status
            New status for the records. Only certain status transitions will be allowed.
        compute_priority
            New priority for these records
        compute_tag
            New tag for these records
        comment
            Adds a new comment to these records
        session
            An existing SQLAlchemy session to use. If None, one will be created. If an existing session
            is used, it will be flushed before returning from this function.

        Returns
        -------
        :
            Metadata about the modification/update.
        """

        if len(record_ids) == 0:
            return UpdateMetadata()

        # do all in a single session
        with self.root_socket.optional_session(session) as session:
            if status is not None:
                if status == RecordStatusEnum.waiting:
                    ret = self.root_socket.records.reset(record_ids=record_ids, session=session)
                if status == RecordStatusEnum.cancelled:
                    ret = self.root_socket.records.cancel(record_ids=record_ids, session=session)
                if status == RecordStatusEnum.invalid:
                    ret = self.root_socket.records.invalidate(record_ids=record_ids, session=session)

                # ignore all other statuses

            if compute_tag is not None or compute_priority is not None:
                ret = self.root_socket.records.modify(
                    record_ids,
                    new_compute_tag=compute_tag,
                    new_compute_priority=compute_priority,
                    session=session,
                )

            if comment:
                ret = self.root_socket.records.add_comment(
                    record_ids=record_ids,
                    user_id=user_id,
                    comment=comment,
                    session=session,
                )

            return ret

    def revert_generic(self, record_id: Sequence[int], revert_status: RecordStatusEnum) -> UpdateMetadata:
        """
        Reverts the status of a record to the previous status

        This dispatches the request to the proper function based on the status. For example,
        if `revert_status` is "deleted", then this will call undelete.
        """

        if revert_status == RecordStatusEnum.cancelled:
            return self.uncancel(record_id)

        if revert_status == RecordStatusEnum.invalid:
            return self.uninvalidate(record_id)

        if revert_status == RecordStatusEnum.deleted:
            return self.undelete(record_id)

        raise RuntimeError(f"Unknown status to revert: ", revert_status)

    def get_waiting_reason(self, record_id: int, *, session: Optional[Session] = None) -> Dict[str, Any]:
        """
        Determines why a record/task is waiting

        This function takes the record ID, not the task ID
        """

        # For getting the record and task info
        rec_stmt = select(
            BaseRecordORM.status, BaseRecordORM.is_service, TaskQueueORM.compute_tag, TaskQueueORM.required_programs
        )
        rec_stmt = rec_stmt.join(TaskQueueORM, TaskQueueORM.record_id == BaseRecordORM.id, isouter=True)
        rec_stmt = rec_stmt.where(BaseRecordORM.id == record_id)

        # All active managers
        manager_stmt = select(ComputeManagerORM.name, ComputeManagerORM.compute_tags, ComputeManagerORM.programs)
        manager_stmt = manager_stmt.where(ComputeManagerORM.status == ManagerStatusEnum.active)

        with self.root_socket.optional_session(session, True) as session:
            rec = session.execute(rec_stmt).one_or_none()

            if rec is None:
                return {"reason": "Record does not exist"}

            rec_status, rec_isservice, rec_tag, rec_programs = rec

            if rec_isservice is True:
                return {"reason": "Record is a service"}

            if rec_status != RecordStatusEnum.waiting:
                return {"reason": "Record is not waiting"}

            if rec_tag is None or rec_programs is None:
                return {"reason": "Missing task? This is a developer error"}

            # Record is waiting. Test each active manager
            managers = session.execute(manager_stmt).all()

            if len(managers) == 0:
                return {"reason": "No active managers"}

            rec_programs = set(rec_programs)

            ret = {"details": {}, "reason": "No manager matches programs & tags"}
            for m_name, m_tags, m_programs in managers:
                missing_programs = rec_programs - m_programs.keys()
                if missing_programs:
                    ret["details"][m_name] = f"Manager missing programs: {missing_programs}"
                elif rec_tag not in m_tags and "*" not in m_tags:
                    ret["details"][m_name] = f'Manager does not handle tag "{rec_tag}"'
                else:
                    ret["details"][m_name] = f"Manager is busy"
                    ret["reason"] = f"Waiting for a free manager"

            return ret



from __future__ import annotations

from typing import TYPE_CHECKING

from sqlalchemy import (
    Column,
    String,
    Integer,
    ForeignKey,
    UniqueConstraint,
    Index,
    CheckConstraint,
    event,
    DDL,
    select,
)
from sqlalchemy.dialects.postgresql import JSONB, ARRAY
from sqlalchemy.orm import relationship
from sqlalchemy.orm.collections import attribute_keyed_dict

from qcfractal.components.molecules.db_models import MoleculeORM
from qcfractal.components.record_db_models import BaseRecordORM
from qcfractal.components.singlepoint.record_db_models import SinglepointRecordORM, QCSpecificationORM
from qcfractal.db_socket import BaseORM

if TYPE_CHECKING:
    from typing import Dict, Any, Optional, Iterable


class ManybodyClusterORM(BaseORM):
    """
    Table for storing fragment clusters in manybody calculations
    """

    __tablename__ = "manybody_cluster"

    id = Column(Integer, primary_key=True)
    manybody_id = Column(Integer, ForeignKey("manybody_record.id", ondelete="cascade"))
    molecule_id = Column(Integer, ForeignKey("molecule.id"), nullable=False)
    mc_level = Column(String, nullable=False)
    fragments = Column(ARRAY(Integer), nullable=False)
    basis = Column(ARRAY(Integer), nullable=False)

    singlepoint_id = Column(Integer, ForeignKey(SinglepointRecordORM.id), nullable=True)

    molecule = relationship(MoleculeORM)
    singlepoint_record = relationship(SinglepointRecordORM)

    __table_args__ = (
        CheckConstraint("array_length(fragments, 1) > 0", name="ck_manybody_cluster_fragments"),
        CheckConstraint("array_length(basis, 1) > 0", name="ck_manybody_cluster_basis"),
        UniqueConstraint("manybody_id", "mc_level", "fragments", "basis", name="ux_manybody_cluster_unique"),
        Index("ix_manybody_cluster_molecule_id", "molecule_id"),
        Index("ix_manybody_cluster_singlepoint_id", "singlepoint_id"),
    )

    _qcportal_model_excludes = ["manybody_id", "id"]


class ManybodySpecificationORM(BaseORM):
    """
    Table for storing manybody specifications
    """

    __tablename__ = "manybody_specification"

    id = Column(Integer, primary_key=True)
    specification_hash = Column(String, nullable=False)

    program = Column(String, nullable=False)
    bsse_correction = Column(ARRAY(String), nullable=False)

    keywords = Column(JSONB, nullable=False)
    protocols = Column(JSONB, nullable=False)

    levels = relationship(
        "ManybodySpecificationLevelsORM", lazy="selectin", collection_class=attribute_keyed_dict("level")
    )

    # Note - specification_hash will not be unique because of the different levels!
    # The levels are stored in another table with FK to this table, so seemingly
    # duplicate rows in this table could have different rows in the levels table
    __table_args__ = (
        Index("ix_manybody_specification_program", "program"),
        CheckConstraint("program = LOWER(program)", name="ck_manybody_specification_program_lower"),
    )

    _qcportal_model_excludes = ["id", "specification_hash"]

    def model_dict(self, exclude: Optional[Iterable[str]] = None) -> Dict[str, Any]:
        d = BaseORM.model_dict(self, exclude)

        # Levels should just be key -> specification
        # map -1 for levels to 'supersystem'
        d["levels"] = {k if k != -1 else "supersystem": v["singlepoint_specification"] for k, v in d["levels"].items()}

        return d

    @property
    def short_description(self) -> str:
        return f"{self.program}~{sorted(self.levels.keys())}"


class ManybodySpecificationLevelsORM(BaseORM):
    """
    Association table for storing singlepoint specifications that are part of a manybody specification
    """

    __tablename__ = "manybody_specification_levels"

    id = Column(Integer, primary_key=True)

    manybody_specification_id = Column(Integer, ForeignKey(ManybodySpecificationORM.id), nullable=False)

    level = Column(Integer, nullable=False)
    singlepoint_specification_id = Column(Integer, ForeignKey(QCSpecificationORM.id), nullable=False)

    singlepoint_specification = relationship(QCSpecificationORM, lazy="joined")

    __table_args__ = (
        UniqueConstraint("manybody_specification_id", "level", name="ux_manybody_specification_levels_unique"),
        Index("ix_manybody_specifications_levels_manybody_specification_id", "manybody_specification_id"),
        Index("ix_manybody_specifications_levels_singlepoint_specification_id", "singlepoint_specification_id"),
    )

    _qcportal_model_excludes = ["id"]


class ManybodyRecordORM(BaseRecordORM):
    """
    Table for storing manybody calculations
    """

    __tablename__ = "manybody_record"

    id = Column(Integer, ForeignKey(BaseRecordORM.id, ondelete="cascade"), primary_key=True)

    initial_molecule_id = Column(Integer, ForeignKey(MoleculeORM.id), nullable=False)
    specification_id = Column(Integer, ForeignKey(ManybodySpecificationORM.id), nullable=False)

    specification = relationship(ManybodySpecificationORM, lazy="selectin")
    initial_molecule = relationship(MoleculeORM)

    clusters = relationship(ManybodyClusterORM, cascade="all, delete-orphan", passive_deletes=True)

    __mapper_args__ = {
        "polymorphic_identity": "manybody",
    }

    _qcportal_model_excludes = [*BaseRecordORM._qcportal_model_excludes, "specification_id"]

    @property
    def short_description(self) -> str:
        return f'{self.initial_molecule.identifiers["molecular_formula"]} {self.specification.short_description}'


# Delete base record if this record is deleted
_del_baserecord_trigger = DDL(
    """
    CREATE TRIGGER qca_manybody_record_delete_base_tr
    AFTER DELETE ON manybody_record
    FOR EACH ROW EXECUTE PROCEDURE qca_base_record_delete();
    """
)

event.listen(ManybodyRecordORM.__table__, "after_create", _del_baserecord_trigger.execute_if(dialect=("postgresql")))

record_direct_children_select = [
    select(BaseRecordORM.id.label("parent_id"), ManybodyClusterORM.singlepoint_id.label("child_id"))
    .join(ManybodyClusterORM, BaseRecordORM.id == ManybodyClusterORM.manybody_id)
    .where(BaseRecordORM.record_type == "manybody"),
]

record_children_select = record_direct_children_select



"""use views for record parents/children

Revision ID: ee68b4a1c346
Revises: c5a3bed43646
Create Date: 2025-04-15 16:12:00.052936

"""

from alembic import op
import sqlalchemy as sa


# revision identifiers, used by Alembic.
revision = "ee68b4a1c346"
down_revision = "c5a3bed43646"
branch_labels = None
depends_on = None


def upgrade():
    # ### commands auto generated by Alembic - please adjust! ###
    op.execute(
        sa.text(
            """
                create or replace view "public"."record_children_view" as  SELECT base_record.id AS parent_id,
                    optimization_trajectory.singlepoint_id AS child_id
                   FROM (base_record
                     JOIN optimization_trajectory ON ((base_record.id = optimization_trajectory.optimization_id)))
                  WHERE (base_record.record_type = 'optimization')
                UNION
                 SELECT base_record.id AS parent_id,
                    gridoptimization_optimization.optimization_id AS child_id
                   FROM (base_record
                     JOIN gridoptimization_optimization ON ((base_record.id = gridoptimization_optimization.gridoptimization_id)))
                  WHERE (base_record.record_type = 'gridoptimization')
                UNION
                 SELECT base_record.id AS parent_id,
                    optimization_trajectory.singlepoint_id AS child_id
                   FROM ((base_record
                     JOIN gridoptimization_optimization ON ((base_record.id = gridoptimization_optimization.gridoptimization_id)))
                     JOIN optimization_trajectory ON ((optimization_trajectory.optimization_id = gridoptimization_optimization.optimization_id)))
                  WHERE (base_record.record_type = 'gridoptimization')
                UNION
                 SELECT base_record.id AS parent_id,
                    torsiondrive_optimization.optimization_id AS child_id
                   FROM (base_record
                     JOIN torsiondrive_optimization ON ((base_record.id = torsiondrive_optimization.torsiondrive_id)))
                  WHERE (base_record.record_type = 'torsiondrive')
                UNION
                 SELECT base_record.id AS parent_id,
                    optimization_trajectory.singlepoint_id AS child_id
                   FROM ((base_record
                     JOIN torsiondrive_optimization ON ((base_record.id = torsiondrive_optimization.torsiondrive_id)))
                     JOIN optimization_trajectory ON ((optimization_trajectory.optimization_id = torsiondrive_optimization.optimization_id)))
                  WHERE (base_record.record_type = 'torsiondrive')
                UNION
                 SELECT base_record.id AS parent_id,
                    manybody_cluster.singlepoint_id AS child_id
                   FROM (base_record
                     JOIN manybody_cluster ON ((base_record.id = manybody_cluster.manybody_id)))
                  WHERE (base_record.record_type = 'manybody')
                UNION
                 SELECT base_record.id AS parent_id,
                    reaction_component.singlepoint_id AS child_id
                   FROM (base_record
                     JOIN reaction_component ON ((base_record.id = reaction_component.reaction_id)))
                  WHERE ((base_record.record_type = 'reaction') AND (reaction_component.singlepoint_id IS NOT NULL))
                UNION
                 SELECT base_record.id AS parent_id,
                    reaction_component.optimization_id AS child_id
                   FROM (base_record
                     JOIN reaction_component ON ((base_record.id = reaction_component.reaction_id)))
                  WHERE ((base_record.record_type = 'reaction') AND (reaction_component.optimization_id IS NOT NULL))
                UNION
                 SELECT base_record.id AS parent_id,
                    optimization_trajectory.singlepoint_id AS child_id
                   FROM ((base_record
                     JOIN reaction_component ON ((base_record.id = reaction_component.reaction_id)))
                     JOIN optimization_trajectory ON ((optimization_trajectory.optimization_id = reaction_component.optimization_id)))
                  WHERE ((base_record.record_type = 'reaction') AND (reaction_component.optimization_id IS NOT NULL))
                UNION
                 SELECT base_record.id AS parent_id,
                    neb_singlepoints.singlepoint_id AS child_id
                   FROM (base_record
                     JOIN neb_singlepoints ON ((base_record.id = neb_singlepoints.neb_id)))
                  WHERE (base_record.record_type = 'neb')
                UNION
                 SELECT base_record.id AS parent_id,
                    neb_optimizations.optimization_id AS child_id
                   FROM (base_record
                     JOIN neb_optimizations ON ((base_record.id = neb_optimizations.neb_id)))
                  WHERE (base_record.record_type = 'neb')
                UNION
                 SELECT base_record.id AS parent_id,
                    optimization_trajectory.singlepoint_id AS child_id
                   FROM ((base_record
                     JOIN neb_optimizations ON ((base_record.id = neb_optimizations.neb_id)))
                     JOIN optimization_trajectory ON ((optimization_trajectory.optimization_id = neb_optimizations.optimization_id)))
                  WHERE (base_record.record_type = 'neb');
            """
        )
    )

    op.execute(
        sa.text(
            """
                create or replace view "public"."record_direct_children_view" as  SELECT base_record.id AS parent_id,
                    optimization_trajectory.singlepoint_id AS child_id
                   FROM (base_record
                     JOIN optimization_trajectory ON ((base_record.id = optimization_trajectory.optimization_id)))
                  WHERE (base_record.record_type = 'optimization')
                UNION
                 SELECT base_record.id AS parent_id,
                    gridoptimization_optimization.optimization_id AS child_id
                   FROM (base_record
                     JOIN gridoptimization_optimization ON ((base_record.id = gridoptimization_optimization.gridoptimization_id)))
                  WHERE (base_record.record_type = 'gridoptimization')
                UNION
                 SELECT base_record.id AS parent_id,
                    torsiondrive_optimization.optimization_id AS child_id
                   FROM (base_record
                     JOIN torsiondrive_optimization ON ((base_record.id = torsiondrive_optimization.torsiondrive_id)))
                  WHERE (base_record.record_type = 'torsiondrive')
                UNION
                 SELECT base_record.id AS parent_id,
                    manybody_cluster.singlepoint_id AS child_id
                   FROM (base_record
                     JOIN manybody_cluster ON ((base_record.id = manybody_cluster.manybody_id)))
                  WHERE (base_record.record_type = 'manybody')
                UNION
                 SELECT base_record.id AS parent_id,
                    reaction_component.singlepoint_id AS child_id
                   FROM (base_record
                     JOIN reaction_component ON ((base_record.id = reaction_component.reaction_id)))
                  WHERE ((base_record.record_type = 'reaction') AND (reaction_component.singlepoint_id IS NOT NULL))
                UNION
                 SELECT base_record.id AS parent_id,
                    reaction_component.optimization_id AS child_id
                   FROM (base_record
                     JOIN reaction_component ON ((base_record.id = reaction_component.reaction_id)))
                  WHERE ((base_record.record_type = 'reaction') AND (reaction_component.optimization_id IS NOT NULL))
                UNION
                 SELECT base_record.id AS parent_id,
                    neb_singlepoints.singlepoint_id AS child_id
                   FROM (base_record
                     JOIN neb_singlepoints ON ((base_record.id = neb_singlepoints.neb_id)))
                  WHERE (base_record.record_type = 'neb')
                UNION
                 SELECT base_record.id AS parent_id,
                    neb_optimizations.optimization_id AS child_id
                   FROM (base_record
                     JOIN neb_optimizations ON ((base_record.id = neb_optimizations.neb_id)))
                  WHERE (base_record.record_type = 'neb');
            """
        )
    )
    # ### end Alembic commands ###


def downgrade():
    # ### commands auto generated by Alembic - please adjust! ###
    op.execute(sa.text("""drop view record_direct_children_view;"""))
    op.execute(sa.text("""drop view record_children_view;"""))
    # ### end Alembic commands ###



from typing import Dict, Set

GLOBAL_ROLE_PERMISSIONS: Dict[str, Dict[str, Set[str]]] = {
    "admin": {
        "*": {"*"},
    },
    "maintain": {
        "information": {"read", "modify"},
        "me": {"read", "modify"},
        "users": {"read", "modify"},
        "groups": {"*"},
        "access_log": {"*"},
        "server_errors": {"*"},
        "internal_jobs": {"*"},
        "managers": {"read"},
        "records": {"*"},
        "datasets": {"*"},
        "projects": {"*"},
    },
    "monitor": {
        "information": {"read"},
        "me": {"read", "modify"},
        "users": {"read", "modify"},  # TODO - DEPRECATE
        "managers": {"read"},
        "records": {"read"},
        "datasets": {"read"},
        "projects": {"read"},
        "access_log": {"read"},
        "server_errors": {"read"},
        "internal_jobs": {"read"},
    },
    "submit": {
        "information": {"read"},
        "me": {"read", "modify"},
        "users": {"read", "modify"},  # TODO - DEPRECATE
        "managers": {"read"},
        "records": {"read", "add", "modify", "delete"},
        "datasets": {"read", "add", "modify", "delete", "create_view"},
        "projects": {"read", "add", "modify", "delete"},
    },
    "read": {
        "information": {"read"},
        "me": {"read", "modify"},
        "users": {"read", "modify"},  # TODO - DEPRECATE
        "managers": {"read"},
        "records": {"read"},
        "datasets": {"read"},
        "projects": {"read"},
    },
    "anonymous": {
        "information": {"read"},
        "managers": {"read"},
        "records": {"read"},
        "datasets": {"read"},
        "projects": {"read"},
    },
    "compute": {
        "information": {"read"},
        "me": {"read", "modify"},
        "managers": {"add", "modify"},
        "tasks": {"claim", "return"},
        "users": {"read", "modify"},  # TODO - DEPRECATE
    },
}



import os

import pytest
from sqlalchemy import create_engine

from qcfractal.config import DatabaseConfig
from qcfractal.db_socket import SQLAlchemySocket
from qcfractal.port_util import find_open_port
from qcfractal.postgres_harness import PostgresHarness


@pytest.mark.parametrize(
    "host,port,username,password,dbname,mdbname,query",
    [
        (
            "localhost",
            5432,
            "test_user_1",
            "test_pass_1234",
            "testing_db_1",
            "testing_db_maint_1",
            {"connect_timeout": 10},
        ),
        (
            "192.168.1.234",
            6543,
            "test_user_1",
            "test_pass_2",
            "testing_db_2",
            "testing_db_maint_2",
            {"connect_timeout": 10},
        ),
        (
            "/var/run/postgresql",
            9876,
            "test_user_4",
            "test_pass_9999",
            "testing_db_4",
            "testing_db_maint_4",
            {"connect_timeout": 10},
        ),
    ],
)
def test_db_connection_uri_convert(host, port, username, password, dbname, mdbname, query):
    # Test all the different URI formats
    db_config = DatabaseConfig(
        data_directory="/tmp/fakedir",  # not used in this test
        base_folder="/tmp/fakedir",  # not used in this test
        own=True,
        host=host,
        port=port,
        username=username,
        password=password,
        database_name=dbname,
        query=query,
        maintenance_db=mdbname,
    )

    # print(db_config.database_uri)
    # print(db_config.safe_uri)
    # print(db_config.sqlalchemy_url.render_as_string(False))
    # print(db_config.sqlalchemy_url.render_as_string(True))
    # print(db_config.psycopg2_dsn)
    # print(db_config.psycopg2_maintenance_dsn)

    common = [host, str(port)]
    if username:
        common.append(username)
    if query:
        for k, v in query.items():
            common.append(f"{k}={v}")

    assert host in db_config.database_uri
    assert host in db_config.psycopg2_dsn
    assert host in db_config.psycopg2_maintenance_dsn
    assert host in db_config.safe_uri

    # SQLAlchemy render_as_string mangles sockets :(
    if not host.startswith("/"):
        assert host in db_config.sqlalchemy_url.render_as_string(False)
        assert host in db_config.sqlalchemy_url.render_as_string(True)

    assert str(port) in db_config.database_uri
    assert str(port) in db_config.psycopg2_dsn
    assert str(port) in db_config.psycopg2_maintenance_dsn
    assert str(port) in db_config.safe_uri
    assert str(port) in db_config.sqlalchemy_url.render_as_string(False)
    assert str(port) in db_config.sqlalchemy_url.render_as_string(True)

    if username:
        assert username in db_config.database_uri
        assert username in db_config.psycopg2_dsn
        assert username in db_config.psycopg2_maintenance_dsn
        assert username in db_config.safe_uri
        assert username in db_config.sqlalchemy_url.render_as_string(False)
        assert username in db_config.sqlalchemy_url.render_as_string(True)

    # Passwords in all but safe_uri
    if password:
        assert password in db_config.database_uri
        assert password in db_config.psycopg2_dsn
        assert password in db_config.psycopg2_maintenance_dsn
        assert password not in db_config.safe_uri
        assert password in db_config.sqlalchemy_url.render_as_string(False)
        assert password not in db_config.sqlalchemy_url.render_as_string(True)

    # dbname in all but maintenance uri
    assert dbname in db_config.database_uri
    assert dbname in db_config.psycopg2_dsn
    assert dbname not in db_config.psycopg2_maintenance_dsn
    assert dbname in db_config.sqlalchemy_url.render_as_string(False)
    assert dbname in db_config.sqlalchemy_url.render_as_string(True)
    assert dbname in db_config.safe_uri

    # Maintenance db only in maintenance uri
    assert mdbname not in db_config.database_uri
    assert mdbname not in db_config.psycopg2_dsn
    assert mdbname in db_config.psycopg2_maintenance_dsn
    assert mdbname not in db_config.sqlalchemy_url.render_as_string(False)
    assert mdbname not in db_config.sqlalchemy_url.render_as_string(True)
    assert mdbname not in db_config.safe_uri

    for k, v in query.items():
        s = f"{k}={v}"
        assert s in db_config.database_uri
        assert s in db_config.psycopg2_dsn
        assert s in db_config.psycopg2_maintenance_dsn
        assert s in db_config.safe_uri
        assert s in db_config.sqlalchemy_url.render_as_string(False)
        assert s in db_config.sqlalchemy_url.render_as_string(True)


def test_db_connection_hosts(tmp_path):
    base_path = tmp_path / "basefolder"
    db_path = tmp_path / "db_data"
    base_path.mkdir()
    db_path.mkdir()

    port = find_open_port()
    db_config = DatabaseConfig(
        port=port,
        data_directory=str(db_path),
        base_folder=str(base_path),
        username="test_connstr_user",
        password="test_connstr_password_1234",
        database_name="testing_db_connstr",
        query={"connect_timeout": 10},
        own=True,
    )

    pg_harness = PostgresHarness(db_config)

    # Change trust method so we can actually check passwords
    pg_harness.initialize_postgres()
    pg_harness.create_database(create_tables=True)
    assert pg_harness.can_connect()

    try:
        # Make sure tests can fail
        new_db_config = pg_harness.config.model_copy(update={"password": "not_correct"})
        new_pg_harness = PostgresHarness(new_db_config)
        assert new_pg_harness.can_connect() is False

        sock_path = os.path.join(db_config.data_directory, "sock")
        for test_host in ["localhost", "127.0.0.1", sock_path]:
            new_db_config = db_config.model_copy(update={"host": test_host})
            assert PostgresHarness(new_db_config).can_connect()
            assert create_engine(new_db_config.sqlalchemy_url).connect()
            SQLAlchemySocket.upgrade_database(new_db_config)
    finally:
        pg_harness.shutdown()


def test_db_connection_full_uri(tmp_path):
    base_path = tmp_path / "basefolder"
    db_path = tmp_path / "db_data"
    base_path.mkdir()
    db_path.mkdir()

    port = find_open_port()
    db_config = DatabaseConfig(
        port=port,
        data_directory=str(db_path),
        base_folder=str(base_path),
        username="test_connstr_user",
        password="test_connstr_password_1234",
        database_name="testing_db_connstr",
        query={"connect_timeout": 10},
        own=True,
    )

    pg_harness = PostgresHarness(db_config)

    # Change trust method so we can actually check passwords
    pg_harness.initialize_postgres()
    pg_harness.create_database(create_tables=True)
    assert pg_harness.can_connect()

    # Make sure tests can fail
    new_db_config = pg_harness.config.model_copy(update={"password": "not_correct"})
    new_pg_harness = PostgresHarness(new_db_config)
    assert new_pg_harness.can_connect() is False

    def can_connect(full_uri):
        test_cfg = db_config.model_copy(update={"full_uri": full_uri})
        assert PostgresHarness(test_cfg).can_connect()
        assert create_engine(test_cfg.sqlalchemy_url).connect()
        SQLAlchemySocket.upgrade_database(test_cfg)

    username = db_config.username
    password = db_config.password
    dbname = db_config.database_name

    for host in ["localhost", "127.0.0.1"]:
        can_connect(f"postgresql://{username}:{password}@{host}:{port}/{dbname}?connect_timeout=10")
        can_connect(f"postgresql://{username}:{password}@{host}:{port}/?dbname={dbname}&connect_timeout=10")
        can_connect(f"postgresql://{host}:{port}/?dbname={dbname}&password={password}&user={username}")
        can_connect(f"postgresql:///?host={host}&port={port}&dbname={dbname}&password={password}&user={username}")
        can_connect(f"postgresql://:{port}/?host={host}&dbname={dbname}&password={password}&user={username}")

    ## Socket file?
    sock_path = os.path.join(db_config.data_directory, "sock")
    if len(sock_path) < 80:
        can_connect(f"postgresql://{username}:{password}@:{port}/{dbname}?host={sock_path}")



from __future__ import annotations

from typing import TYPE_CHECKING, Tuple, Optional, Dict, Union, Any

try:
    import pydantic.v1 as pydantic
except ImportError:
    import pydantic
from qcelemental.models import Molecule, FailedOperation, ComputeError, AtomicResult as QCEl_AtomicResult

from qcarchivetesting.helpers import read_procedure_data, read_record_data, find_test_data
from qcfractal.components.manybody.record_db_models import ManybodyRecordORM
from qcfractal.testing_helpers import run_service
from qcportal.manybody import ManybodySpecification, ManybodyRecord
from qcportal.singlepoint import QCSpecification, SinglepointProtocols
from qcportal.record_models import PriorityEnum, RecordStatusEnum, RecordTask

if TYPE_CHECKING:
    from qcfractal.db_socket import SQLAlchemySocket
    from qcportal.managers import ManagerName


all_test_data = find_test_data("mb_*")
all_includes = ["initial_molecule", "clusters", "molecule", "comments"]

test_specs = [
    ManybodySpecification(
        program="qcmanybody",
        bsse_correction=["nocp"],
        levels={
            1: QCSpecification(
                program="Prog1",
                driver="energy",
                method="b3lyp",
                basis="6-31G*",
                keywords={"k": "value"},
                protocols=SinglepointProtocols(wavefunction="all"),
            ),
        },
        keywords={"return_total_data": True},
    ),
    ManybodySpecification(
        program="qcmanybody",
        bsse_correction=["cp"],
        levels={
            1: QCSpecification(
                program="Prog2",
                driver="energy",
                method="Hf",
                basis="def2-tzVP",
                keywords={"k": "value"},
                protocols=SinglepointProtocols(wavefunction="all"),
            ),
        },
        keywords={"return_total_data": True},
    ),
    ManybodySpecification(
        program="qcmanybody",
        bsse_correction=["cp", "vmfc"],
        levels={
            1: QCSpecification(
                program="Prog3",
                driver="energy",
                method="mp2",
                basis="sto-3g",
                keywords={"k": "v"},
            ),
            "supersystem": QCSpecification(
                program="Prog3",
                driver="energy",
                method="Hf",
                basis="sto-3g",
                keywords={"k": "v"},
            ),
        },
        keywords={"return_total_data": False},
    ),
]


def compare_manybody_specs(
    input_spec: Union[ManybodySpecification, Dict[str, Any]],
    output_spec: Union[ManybodySpecification, Dict[str, Any]],
) -> bool:
    if isinstance(input_spec, dict):
        input_spec = ManybodySpecification(**input_spec)
    if isinstance(output_spec, dict):
        output_spec = ManybodySpecification(**output_spec)

    return input_spec == output_spec


def generate_task_key(task: RecordTask):
    # task is a singlepoint
    inp_data = task.function_kwargs["input_data"]
    assert inp_data["schema_name"] == "qcschema_input"

    mol_hash = inp_data["molecule"]["identifiers"]["molecule_hash"]
    return "singlepoint" + "|" + mol_hash


def load_procedure_data(name: str) -> Tuple[ManybodySpecification, Molecule, Dict[str, QCEl_AtomicResult]]:
    data = read_procedure_data(name)

    return (
        pydantic.parse_obj_as(ManybodySpecification, data["specification"]),
        pydantic.parse_obj_as(Molecule, data["molecule"]),
        pydantic.parse_obj_as(Dict[str, QCEl_AtomicResult], data["results"]),
    )


def load_record_data(name: str) -> ManybodyRecord:
    data = read_record_data(name)
    return ManybodyRecord(**data)


def submit_procedure_data(
    storage_socket: SQLAlchemySocket,
    name: str,
    compute_tag: Optional[str] = "*",
    compute_priority: PriorityEnum = PriorityEnum.normal,
) -> Tuple[int, Dict[str, QCEl_AtomicResult]]:
    input_spec, molecule, result = load_procedure_data(name)
    meta, record_ids = storage_socket.records.manybody.add(
        [molecule], input_spec, compute_tag, compute_priority, None, True
    )
    assert meta.success
    assert len(record_ids) == 1
    assert meta.n_inserted == 1

    return record_ids[0], result


def run_procedure_data(
    storage_socket: SQLAlchemySocket,
    manager_name: ManagerName,
    name: str,
    compute_tag: Optional[str] = "*",
    compute_priority: PriorityEnum = PriorityEnum.normal,
    end_status: RecordStatusEnum = RecordStatusEnum.complete,
):
    record_id, result = submit_procedure_data(storage_socket, name, compute_tag, compute_priority)

    with storage_socket.session_scope() as session:
        record = session.get(ManybodyRecordORM, record_id)
        assert record.status == RecordStatusEnum.waiting

    if end_status == RecordStatusEnum.error:
        failed_op = FailedOperation(
            error=ComputeError(error_type="test_error", error_message="this is just a test error"),
        )
        result = {x: failed_op for x in result}

    finished, n_optimizations = run_service(storage_socket, manager_name, record_id, generate_task_key, result, 200)
    assert finished

    with storage_socket.session_scope() as session:
        record = session.get(ManybodyRecordORM, record_id)
        assert record.status == end_status

    return record_id



from __future__ import annotations

from typing import Optional, Iterable, Dict, Any

from sqlalchemy import (
    Column,
    Integer,
    String,
    JSON,
    Index,
    Computed,
    ForeignKey,
    UniqueConstraint,
    Enum,
)
from sqlalchemy.orm import relationship
from sqlalchemy.orm.collections import attribute_keyed_dict

from qcfractal.components.auth.db_models import UserIDMapSubquery, UserORM
from qcfractal.components.external_files.db_models import ExternalFileORM
from qcfractal.components.internal_jobs.db_models import InternalJobORM
from qcfractal.db_socket import BaseORM, MsgpackExt
from qcportal.dataset_models import DatasetAttachmentType


class BaseDatasetORM(BaseORM):
    """
    A base class for all dataset ORM
    """

    __tablename__ = "base_dataset"

    id = Column(Integer, primary_key=True)
    dataset_type = Column(String, nullable=False)

    lname = Column(String(100), Computed("LOWER(name)"), nullable=False)
    name = Column(String(100), nullable=False)

    tags = Column(JSON, nullable=False)
    tagline = Column(String, nullable=False)
    description = Column(String, nullable=False)

    # Who created this dataset
    creator_user_id = Column(Integer, ForeignKey(UserORM.id), nullable=True)

    creator_user = relationship(
        UserIDMapSubquery,
        foreign_keys=[creator_user_id],
        primaryjoin="BaseDatasetORM.creator_user_id == UserIDMapSubquery.id",
        lazy="selectin",
        viewonly=True,
    )

    default_compute_tag = Column(String, nullable=False)
    default_compute_priority = Column(Integer, nullable=False)

    provenance = Column(JSON, nullable=False)
    extras = Column("extras", JSON, nullable=False)

    contributed_values = relationship(
        "ContributedValuesORM",
        collection_class=attribute_keyed_dict("name"),
        cascade="all, delete-orphan",
        passive_deletes=True,
    )

    attachments = relationship(
        "DatasetAttachmentORM",
        cascade="all, delete-orphan",
        passive_deletes=True,
    )

    __table_args__ = (
        UniqueConstraint("dataset_type", "lname", name="ux_base_dataset_dataset_type_lname"),
        Index("ix_base_dataset_dataset_type", "dataset_type"),
        Index("ix_base_dataset_creator_user_id", "creator_user_id"),
    )

    __mapper_args__ = {"polymorphic_on": "dataset_type"}

    _qcportal_model_excludes = ["lname", "creator_user_id"]

    def model_dict(self, exclude: Optional[Iterable[str]] = None) -> Dict[str, Any]:
        d = BaseORM.model_dict(self, exclude)

        # TODO - DEPRECATED - REMOVE EVENTUALLY
        d["group"] = "default"
        d["visibility"] = True
        if "extras" in d:
            d["metadata"] = d.pop("extras")
            d["extras"] = {}
        if "default_compute_tag" in d:
            d["default_tag"] = d.pop("default_compute_tag")
        if "default_compute_priority" in d:
            d["default_priority"] = d.pop("default_compute_priority")

        d["owner_user"] = self.creator_user.username if self.creator_user is not None else None
        d.pop("creator_user", None)
        d["owner_group"] = None

        return d


class ContributedValuesORM(BaseORM):
    __tablename__ = "contributed_values"

    dataset_id = Column(Integer, ForeignKey("base_dataset.id", ondelete="cascade"), primary_key=True)

    name = Column(String, nullable=False, primary_key=True)
    values = Column(MsgpackExt, nullable=False)
    index = Column(MsgpackExt, nullable=False)
    values_structure = Column(JSON, nullable=False)

    theory_level = Column(JSON, nullable=False)
    units = Column(String, nullable=False)
    theory_level_details = Column(JSON)

    citations = Column(JSON)
    external_url = Column(String)
    doi = Column(String)

    comments = Column(String)

    __table_args__ = (Index("ix_contributed_values_dataset_id", "dataset_id"),)

    _qcportal_model_excludes = ["dataset_id"]


class DatasetInternalJobORM(BaseORM):
    __tablename__ = "dataset_internal_job"

    internal_job_id = Column(Integer, ForeignKey(InternalJobORM.id, ondelete="cascade"), primary_key=True)
    dataset_id = Column(Integer, ForeignKey("base_dataset.id", ondelete="cascade"), primary_key=True)


class DatasetRecordCountORM(BaseORM):
    # We don't include this table as a relationship in the DatasetORM. We only
    # use it when we list datasets
    __tablename__ = "dataset_record_count"

    dataset_id = Column(Integer, ForeignKey("base_dataset.id", ondelete="cascade"), primary_key=True)
    record_count = Column(Integer, nullable=False, default=0)


class DatasetAttachmentORM(ExternalFileORM):
    __tablename__ = "dataset_attachment"

    id = Column(Integer, ForeignKey(ExternalFileORM.id, ondelete="cascade"), primary_key=True)
    dataset_id = Column(Integer, ForeignKey("base_dataset.id", ondelete="cascade"), nullable=False)

    attachment_type = Column(Enum(DatasetAttachmentType), nullable=False)

    __mapper_args__ = {"polymorphic_identity": "dataset_attachment"}

    __table_args__ = (Index("ix_dataset_attachment_dataset_id", "dataset_id"),)

    _qcportal_model_excludes = ["dataset_id"]



from __future__ import annotations

from flask import session, g
from flask_jwt_extended import verify_jwt_in_request, get_jwt_identity, get_jwt
from jwt.exceptions import ExpiredSignatureError
from werkzeug.exceptions import InternalServerError

from qcfractal.flask_app import storage_socket
from qcportal.exceptions import AuthorizationFailure, AuthenticationFailure
from qcportal.utils import time_based_cache


@time_based_cache(seconds=3, maxsize=256)
def _cached_verify(user_id: int):
    return storage_socket.auth.verify(user_id=user_id)


def load_logged_in_user():
    ##############################################
    # Load any user information from a JWT or
    # session info in the database (retrieved
    # via the typical flask session mechanism)
    ##############################################
    user_id = None
    username = None
    role = None
    groups = []

    try:
        # Is the info stored in the session?
        if session and "user_id" in session:
            user_id = int(session["user_id"])  # may be a string? Just to make sure

            user_info = _cached_verify(user_id=user_id)
            username = user_info.username
            role = user_info.role
            groups = user_info.groups
        elif verify_jwt_in_request(optional=True) is not None:
            user_id = get_jwt_identity()

            if user_id is not None:
                # user_id is stored in the JWT as a string
                user_id = int(user_id)

                # Get from JWT in header
                # TODO - some of these may not be None in the future
                claims = get_jwt()
                username = claims.get("username", None)
                role = claims.get("role", None)
                groups = claims.get("groups", [])
    except (AuthorizationFailure, AuthenticationFailure):
        raise
    except ExpiredSignatureError as e:
        raise AuthenticationFailure(f"Authentication failure - JWT Token has expired: {str(e)}")
    except Exception as e:
        raise InternalServerError(f"Failed to verify user info: {str(e)}")

    # Store the user in the global app/request context
    g.user_id = user_id
    g.username = username
    g.role = role
    g.groups = groups



"""
Additional column types for SQLAlchemy
"""

try:
    from pydantic.v1 import BaseModel
    from pydantic.v1.json import pydantic_encoder
except ImportError:
    from pydantic import BaseModel
    from pydantic.json import pydantic_encoder

from typing import Any

import msgpack
import numpy as np
from sqlalchemy import TypeDecorator
from sqlalchemy.dialects.postgresql import BYTEA


def _msgpackext_encode(obj: Any) -> Any:
    # First try pydantic base objects
    try:
        return pydantic_encoder(obj)
    except TypeError:
        pass

    if isinstance(obj, np.ndarray):
        if obj.shape:
            data = {b"_nd_": True, b"dtype": obj.dtype.str, b"data": np.ascontiguousarray(obj).tobytes()}
            if len(obj.shape) > 1:
                data[b"shape"] = obj.shape
            return data

        else:
            # Converts np.array(5) -> 5
            return obj.tolist()

    return obj


def _msgpackext_decode(obj: Any) -> Any:
    if b"_nd_" in obj:
        arr = np.frombuffer(obj[b"data"], dtype=obj[b"dtype"])
        if b"shape" in obj:
            arr.shape = obj[b"shape"]

        return arr

    return obj


class MsgpackExt(TypeDecorator):
    """Converts JSON-like data to msgpack with full NumPy Array support."""

    impl = BYTEA

    # I believe caching is only used when, for example, you filter by a column. But we
    # shouldn't ever do that with msgpack
    cache_ok = False

    def process_bind_param(self, value, dialect):
        if value is None:
            return value
        else:
            return msgpack.dumps(value, default=_msgpackext_encode, use_bin_type=True)

    def process_result_value(self, value, dialect):
        if value is None:
            return value
        else:
            return msgpack.loads(value, object_hook=_msgpackext_decode, raw=False)


class PlainMsgpackExt(TypeDecorator):
    """Converts JSON-like data to msgpack using standard msgpack

    This does not support NumPy"""

    impl = BYTEA

    # I believe caching is only used when, for example, you filter by a column. But we
    # shouldn't ever do that with msgpack
    cache_ok = False

    def process_bind_param(self, value, dialect):
        if value is None:
            return value
        else:
            return msgpack.dumps(value)

    def process_result_value(self, value, dialect):
        if value is None:
            return value
        else:
            return msgpack.loads(value)



"""Add internal job serial group

Revision ID: 3690c677f8d1
Revises: 5f6f804e11d3
Create Date: 2025-01-10 16:08:36.541807

"""

import sqlalchemy as sa
from alembic import op

# revision identifiers, used by Alembic.
revision = "3690c677f8d1"
down_revision = "5f6f804e11d3"
branch_labels = None
depends_on = None


def upgrade():
    # ### commands auto generated by Alembic - please adjust! ###
    op.add_column("internal_jobs", sa.Column("serial_group", sa.String(), nullable=True))
    op.create_index(
        "ux_internal_jobs_status_serial_group",
        "internal_jobs",
        ["status", "serial_group"],
        unique=True,
        postgresql_where=sa.text("status = 'running'"),
    )
    # ### end Alembic commands ###


def downgrade():
    # ### commands auto generated by Alembic - please adjust! ###
    op.drop_index(
        "ux_internal_jobs_status_serial_group",
        table_name="internal_jobs",
        postgresql_where=sa.text("status = 'running'"),
    )
    op.drop_column("internal_jobs", "serial_group")
    # ### end Alembic commands ###



from __future__ import annotations

from typing import TYPE_CHECKING

import pytest

from qcarchivetesting import load_molecule_data, test_users
from qcfractal.components.testing_helpers import convert_to_plain_qcschema_result
from qcportal.managers import ManagerName
from qcportal.molecules import Molecule
from qcportal.record_models import RecordStatusEnum, PriorityEnum, RecordTask
from qcportal.singlepoint import QCSpecification, SinglepointDriver, SinglepointProtocols, compare_singlepoint_records
from qcportal.utils import now_at_utc
from .record_db_models import SinglepointRecordORM
from .testing_helpers import test_specs, load_procedure_data, load_record_data, run_procedure_data
from ..record_utils import build_extras_properties

if TYPE_CHECKING:
    from qcfractal.db_socket import SQLAlchemySocket
    from qcarchivetesting.testing_classes import QCATestingSnowflake
    from sqlalchemy.orm.session import Session
    from typing import Dict, List


def coalesce(x):
    if x is None:
        return ""
    return x


def _compare_record_with_schema(record_orm, result_schema):
    assert record_orm.status == RecordStatusEnum.complete
    assert record_orm.specification.program == result_schema.provenance.creator.lower()

    assert record_orm.specification.driver == result_schema.driver
    assert record_orm.specification.method == result_schema.model.method

    # some task specs still return NULL/None for basis
    assert coalesce(record_orm.specification.basis) == coalesce(result_schema.model.basis)

    assert record_orm.specification.keywords == result_schema.keywords
    assert record_orm.specification.protocols == result_schema.protocols

    assert len(record_orm.compute_history) == 1
    assert record_orm.compute_history[0].status == RecordStatusEnum.complete

    assert record_orm.compute_history[0].provenance == result_schema.provenance

    # Use plain schema, where compressed stuff is removed
    new_extras, new_properties = build_extras_properties(result_schema.copy(deep=True))
    assert record_orm.properties == new_properties
    assert record_orm.extras == new_extras

    # do some common properties themselves
    result_dict = result_schema.dict(include={"return_result"}, encoding="json")
    assert record_orm.properties.get("nuclear_repulsion_energy") == result_schema.properties.nuclear_repulsion_energy
    assert record_orm.properties.get("return_energy") == result_schema.properties.return_energy
    assert record_orm.properties.get("scf_iterations") == result_schema.properties.scf_iterations
    assert record_orm.properties.get("scf_total_energy") == result_schema.properties.scf_total_energy
    assert record_orm.properties.get("return_result") == result_dict["return_result"]

    if result_schema.wavefunction is not None:
        wfn_1 = record_orm.wavefunction.get_wavefunction().dict(encoding="json")
        wfn_2 = record_orm.wavefunction.get_wavefunction().dict(encoding="json")
        wfn_ref = result_schema.wavefunction.dict(encoding="json")
        assert wfn_1 == wfn_2 == wfn_ref
    else:
        assert record_orm.wavefunction is None

    if result_schema.native_files is not None:
        assert record_orm.native_files.keys() == result_schema.native_files.keys()

        for k, v in result_schema.native_files.items():
            assert record_orm.native_files[k].get_file() == v
    else:
        assert record_orm.native_files is None

    for k in ("stdout", "stderr", "error"):
        plain_output = getattr(result_schema, k)
        if plain_output is not None:
            out_str = record_orm.compute_history[0].outputs[k].get_output()
            assert out_str == plain_output
        else:
            assert k not in record_orm.compute_history[0].outputs


@pytest.mark.parametrize("spec", test_specs)
def test_singlepoint_socket_task_spec(
    storage_socket: SQLAlchemySocket,
    spec: QCSpecification,
    activated_manager_name: ManagerName,
    activated_manager_programs: Dict[str, List[str]],
):
    water = load_molecule_data("water_dimer_minima")
    hooh = load_molecule_data("hooh")
    ne4 = load_molecule_data("neon_tetramer")
    all_mols = [water, hooh, ne4]

    time_0 = now_at_utc()
    meta, id = storage_socket.records.singlepoint.add(all_mols, spec, "tag1", PriorityEnum.low, None, True)
    time_1 = now_at_utc()
    assert meta.success

    tasks = storage_socket.tasks.claim_tasks(activated_manager_name.fullname, activated_manager_programs, ["*"])
    tasks = [RecordTask(**t) for t in tasks]

    assert len(tasks) == 3
    for t in tasks:
        function_kwargs = t.function_kwargs
        assert function_kwargs["input_data"]["model"] == {"method": spec.method, "basis": spec.basis}
        assert function_kwargs["input_data"]["protocols"] == spec.protocols.dict(
            exclude_defaults=True, exclude_unset=True
        )
        assert function_kwargs["input_data"]["keywords"] == spec.keywords
        assert function_kwargs["program"] == spec.program
        assert t.compute_tag == "tag1"
        assert t.compute_priority == PriorityEnum.low

    rec_id_mol_map = {
        id[0]: all_mols[0],
        id[1]: all_mols[1],
        id[2]: all_mols[2],
    }

    assert Molecule(**tasks[0].function_kwargs["input_data"]["molecule"]) == rec_id_mol_map[tasks[0].record_id]
    assert Molecule(**tasks[1].function_kwargs["input_data"]["molecule"]) == rec_id_mol_map[tasks[1].record_id]
    assert Molecule(**tasks[2].function_kwargs["input_data"]["molecule"]) == rec_id_mol_map[tasks[2].record_id]


def test_singlepoint_socket_find_existing_1(storage_socket: SQLAlchemySocket):
    spec = QCSpecification(
        program="prog1",
        driver=SinglepointDriver.energy,
        method="b3lyp",
        basis="6-31G*",
        keywords={"k": "value"},
        protocols=SinglepointProtocols(wavefunction="all"),
    )

    water = load_molecule_data("water_dimer_minima")
    meta, id1 = storage_socket.records.singlepoint.add([water], spec, "*", PriorityEnum.normal, None, True)
    assert meta.n_inserted == 1
    assert meta.inserted_idx == [0]

    meta, id2 = storage_socket.records.singlepoint.add([water], spec, "*", PriorityEnum.normal, None, True)
    assert meta.n_inserted == 0
    assert meta.n_existing == 1
    assert meta.existing_idx == [0]
    assert id1 == id2


def test_singlepoint_socket_find_existing_2(storage_socket: SQLAlchemySocket):
    # Test case sensitivity
    spec1 = QCSpecification(
        program="prog1",
        driver=SinglepointDriver.energy,
        method="b3lyp",
        basis="6-31G*",
        keywords={"k": "value"},
        protocols=SinglepointProtocols(wavefunction="all"),
    )

    spec2 = QCSpecification(
        program="pRog1",
        driver=SinglepointDriver.energy,
        method="b3lYp",
        basis="6-31g*",
        keywords={"k": "value"},
        protocols=SinglepointProtocols(wavefunction="all"),
    )

    water = load_molecule_data("water_dimer_minima")
    meta, id1 = storage_socket.records.singlepoint.add([water], spec1, "*", PriorityEnum.normal, None, True)
    assert meta.n_inserted == 1
    assert meta.inserted_idx == [0]

    meta, id2 = storage_socket.records.singlepoint.add([water], spec2, "*", PriorityEnum.normal, None, True)
    assert meta.n_inserted == 0
    assert meta.n_existing == 1
    assert meta.existing_idx == [0]
    assert id1 == id2


def test_singlepoint_socket_find_existing_3(storage_socket: SQLAlchemySocket):
    # Test default keywords and protocols
    spec1 = QCSpecification(
        program="prog1",
        driver=SinglepointDriver.energy,
        method="b3lyp",
        basis="6-31G*",
        keywords={},
        protocols=SinglepointProtocols(wavefunction="none"),
    )

    spec2 = QCSpecification(
        program="prog1",
        driver=SinglepointDriver.energy,
        method="b3lyp",
        basis="6-31G*",
    )

    water = load_molecule_data("water_dimer_minima")
    meta, id1 = storage_socket.records.singlepoint.add([water], spec1, "*", PriorityEnum.normal, None, True)
    assert meta.n_inserted == 1
    assert meta.inserted_idx == [0]

    meta, id2 = storage_socket.records.singlepoint.add([water], spec2, "*", PriorityEnum.normal, None, True)
    assert meta.n_inserted == 0
    assert meta.n_existing == 1
    assert meta.existing_idx == [0]
    assert id1 == id2


def test_singlepoint_socket_find_existing_4(storage_socket: SQLAlchemySocket):
    # Test None basis
    spec1 = QCSpecification(
        program="prog1",
        driver=SinglepointDriver.energy,
        method="b3lyp",
        basis=None,
    )

    spec2 = QCSpecification(
        program="prog1",
        driver=SinglepointDriver.energy,
        method="b3lyp",
        basis="",
    )

    water = load_molecule_data("water_dimer_minima")
    meta, id1 = storage_socket.records.singlepoint.add([water], spec1, "*", PriorityEnum.normal, None, True)
    assert meta.n_inserted == 1
    assert meta.inserted_idx == [0]

    meta, id2 = storage_socket.records.singlepoint.add([water], spec2, "*", PriorityEnum.normal, None, True)
    assert meta.n_inserted == 0
    assert meta.n_existing == 1
    assert meta.existing_idx == [0]
    assert id1 == id2


def test_singlepoint_socket_find_existing_5(storage_socket: SQLAlchemySocket):
    # Test adding molecule by id

    water = load_molecule_data("water_dimer_minima")
    kw = {"a": "value"}
    _, mol_ids = storage_socket.molecules.add([water])

    spec1 = QCSpecification(program="prog1", driver=SinglepointDriver.energy, method="b3lyp", basis=None, keywords=kw)

    spec2 = QCSpecification(program="prog1", driver=SinglepointDriver.energy, method="b3lyp", basis="", keywords=kw)

    meta, id1 = storage_socket.records.singlepoint.add([water], spec1, "*", PriorityEnum.normal, None, True)
    assert meta.n_inserted == 1
    assert meta.inserted_idx == [0]

    meta, id2 = storage_socket.records.singlepoint.add(mol_ids, spec2, "*", PriorityEnum.normal, None, True)
    assert meta.n_inserted == 0
    assert meta.n_existing == 1
    assert meta.existing_idx == [0]
    assert id1 == id2


def test_singlepoint_socket_run(
    storage_socket: SQLAlchemySocket, session: Session, activated_manager_name: ManagerName
):
    test_names = [
        "sp_psi4_benzene_energy_1",
        "sp_psi4_peroxide_energy_wfn",
        "sp_rdkit_water_energy",
        "sp_psi4_h2_b3lyp_nativefiles",
    ]

    all_results = []
    all_id = []

    for test_name in test_names:
        _, _, result_data = load_procedure_data(test_name)
        record_id = run_procedure_data(storage_socket, activated_manager_name, test_name)
        all_results.append(result_data)
        all_id.append(record_id)

    for rec_id, result in zip(all_id, all_results):
        record = session.get(SinglepointRecordORM, rec_id)

        plain_result = convert_to_plain_qcschema_result(result)
        _compare_record_with_schema(record, plain_result)

        # some extra stuff
        desc_info = storage_socket.records.get_short_descriptions([rec_id])[0]
        short_desc = desc_info["description"]
        assert desc_info["record_type"] == record.record_type
        assert desc_info["created_on"] == record.created_on
        assert record.specification.program in short_desc
        assert record.specification.method in short_desc


def test_singlepoint_socket_insert_full_schema_v1(secure_snowflake: QCATestingSnowflake):
    test_names = [
        "sp_psi4_benzene_energy_1",
        "sp_psi4_benzene_energy_2",
        "sp_psi4_benzene_energy_3",
        "sp_psi4_fluoroethane_wfn",
        "sp_psi4_h2_b3lyp_nativefiles",
        "sp_psi4_peroxide_energy_wfn",
        "sp_psi4_water_energy",
        "sp_psi4_water_gradient",
        "sp_psi4_water_hessian",
        "sp_rdkit_benzene_energy",
        "sp_rdkit_water_energy",
    ]

    storage_socket = secure_snowflake.get_storage_socket()
    client = secure_snowflake.client("submit_user", test_users["submit_user"]["pw"])
    user_id = client.get_user().id

    all_ids = []

    for test_name in test_names:
        _, _, result_schema = load_procedure_data(test_name)

        plain_schema = convert_to_plain_qcschema_result(result_schema)

        # Need a full copy of results - they can get mutated
        with storage_socket.session_scope() as session2:
            ins_ids_1 = storage_socket.records.insert_full_schema_v1(session2, [result_schema.copy(deep=True)], user_id)
            ins_ids_2 = storage_socket.records.insert_full_schema_v1(session2, [plain_schema.copy(deep=True)], user_id)

        ins_id_1 = ins_ids_1[0]
        ins_id_2 = ins_ids_2[0]

        # insert_full_schema always inserts
        assert ins_id_1 != ins_id_2
        assert ins_id_1 not in all_ids
        assert ins_id_2 not in all_ids
        all_ids.extend([ins_id_1, ins_id_2])

        with storage_socket.session_scope() as session:
            rec_1 = session.get(SinglepointRecordORM, ins_id_1)
            rec_2 = session.get(SinglepointRecordORM, ins_id_2)

            assert rec_1.creator_user_id == user_id
            assert rec_2.creator_user_id == user_id

            _compare_record_with_schema(rec_1, plain_schema)
            _compare_record_with_schema(rec_2, plain_schema)


def test_singlepoint_socket_insert_full_qcportal_record(secure_snowflake: QCATestingSnowflake):
    test_names = [
        "sp_psi4_benzene_energy_1",
        "sp_psi4_peroxide_energy_wfn",
        "sp_rdkit_water_energy",
        "sp_psi4_h2_b3lyp_nativefiles",
        "sp_error_123975361",
        "sp_error_119608646",
    ]

    storage_socket = secure_snowflake.get_storage_socket()
    client = secure_snowflake.client("submit_user", test_users["submit_user"]["pw"])
    user_id = client.get_user().id

    for test_name in test_names:
        initial_record = load_record_data(test_name)
        initial_record_copy = initial_record.copy(deep=True)

        # Need a full copy of results - they can get mutated
        with storage_socket.session_scope() as session:
            ins_ids = storage_socket.records.insert_full_qcportal_records(session, [initial_record_copy], user_id)

        rec_1 = client.get_singlepoints(ins_ids[0], include=["**"])

        compare_singlepoint_records(rec_1, initial_record)

        assert rec_1.creator_user == "submit_user"



from qcarchivetesting import load_hash_test_data
from qcfractal.components.neb.record_db_models import NEBSpecificationORM
from qcfractal.components.testing_fixtures import spec_test_runner
from qcfractal.db_socket import SQLAlchemySocket
from qcportal.neb import NEBSpecification, NEBKeywords
from qcportal.optimization import OptimizationSpecification, OptimizationProtocols
from qcportal.singlepoint import QCSpecification, SinglepointDriver, SinglepointProtocols

common_sp_spec = QCSpecification(
    program="psi4",
    method="CCSD(T)",
    basis="def2-tzvp",
    keywords={"k1": "values1"},
    driver=SinglepointDriver.gradient,
    protocols=SinglepointProtocols(wavefunction="all"),
)

common_opt_spec = OptimizationSpecification(
    program="optprog2",
    keywords={"k": "value"},
    protocols=OptimizationProtocols(trajectory="none"),
    qc_specification=common_sp_spec.copy(),
)


def test_neb_hash_canaries(storage_socket: SQLAlchemySocket):
    # Test data is hash : spec dict
    test_data = load_hash_test_data("neb_specification_tests")

    # Hashes are independent of opt and sp spec
    # TODO - protocols are not part of model
    for t in test_data.values():
        del t["protocols"]
    spec_map = [
        (k, NEBSpecification(singlepoint_specification=common_sp_spec, optimization_specification=common_opt_spec, **v))
        for k, v in test_data.items()
    ]

    specs = [x[1] for x in spec_map]
    meta, ids = storage_socket.records.neb.add_specifications(specs)
    assert meta.success
    assert len(ids) == len(specs)
    assert meta.n_existing == 0

    with storage_socket.session_scope() as session:
        for spec_id, (spec_hash, _) in zip(ids, spec_map):
            spec_orm = session.get(NEBSpecificationORM, spec_id)
            assert spec_orm.specification_hash == spec_hash


def test_neb_socket_add_specification_same_1(spec_test_runner):
    spec1 = NEBSpecification(
        program="geometric",
        keywords=NEBKeywords(images=11, spring_constant=1.0, optimize_ts=True, optimize_endpoints=True),
        singlepoint_specification=common_sp_spec,
    )

    spec_test_runner("neb", spec1, spec1, True)


def test_neb_socket_add_specification_same_2(spec_test_runner):
    # model handling defaults
    spec1 = NEBSpecification(
        program="geometric",
        keywords=NEBKeywords(images=11, spring_constant=1.0, optimize_ts=False),
        singlepoint_specification=common_sp_spec,
    )

    spec2 = NEBSpecification(
        program="geometric",
        keywords=NEBKeywords(),
        singlepoint_specification=common_sp_spec,
    )

    spec_test_runner("neb", spec1, spec2, True)


def test_neb_socket_add_specification_same_3(spec_test_runner):
    # Test keywords defaults
    spec1 = NEBSpecification(
        program="geometric",
        keywords=NEBKeywords(),
        singlepoint_specification=common_sp_spec,
    )

    spec2 = NEBSpecification(
        program="geometric", keywords={}, singlepoint_specification=common_sp_spec, optimization_specification=None
    )

    spec_test_runner("neb", spec1, spec2, True)


def test_neb_socket_add_specification_diff_1(spec_test_runner):
    # Specifying opt spec
    spec1 = NEBSpecification(
        program="geometric",
        keywords=NEBKeywords(
            images=11,
            spring_constant=1.0,
            spring_type=1,
        ),
        singlepoint_specification=common_sp_spec,
    )

    spec2 = spec1.copy(update={"optimization_specification": common_opt_spec})
    spec_test_runner("neb", spec1, spec2, False)


def test_neb_socket_add_specification_diff_2(spec_test_runner):
    # change images
    spec1 = NEBSpecification(
        program="geometric",
        keywords=NEBKeywords(
            images=21,
            spring_constant=1.5,
            optimize_endpoints=False,
            spring_type=0,
        ),
        singlepoint_specification=common_sp_spec,
    )

    neb_kw = spec1.keywords.copy(update={"images": 31})
    spec2 = spec1.copy(update={"keywords": neb_kw})
    spec_test_runner("neb", spec1, spec2, False)


def test_neb_socket_add_specification_diff_3(spec_test_runner):
    # change spring constant
    spec1 = NEBSpecification(
        program="geometric",
        keywords=NEBKeywords(
            images=21,
            spring_constant=1.5,
            optimize_endpoints=False,
            spring_type=0,
        ),
        singlepoint_specification=common_sp_spec,
    )

    neb_kw = spec1.keywords.copy(update={"spring_constant": 1.51})
    spec2 = spec1.copy(update={"keywords": neb_kw})
    spec_test_runner("neb", spec1, spec2, False)


def test_neb_socket_add_specification_diff_4(spec_test_runner):
    # change opt endpoints
    spec1 = NEBSpecification(
        program="geometric",
        keywords=NEBKeywords(
            images=21,
            spring_constant=1.5,
            optimize_endpoints=False,
            spring_type=0,
        ),
        singlepoint_specification=common_sp_spec,
    )

    neb_kw = spec1.keywords.copy(update={"optimize_endpoints": True})
    spec2 = spec1.copy(update={"keywords": neb_kw})
    spec_test_runner("neb", spec1, spec2, False)


def test_neb_socket_add_specification_diff_5(spec_test_runner):
    # change max force
    spec1 = NEBSpecification(
        program="geometric",
        keywords=NEBKeywords(
            images=21,
            spring_constant=1.5,
            optimize_endpoints=False,
            spring_type=0,
        ),
        singlepoint_specification=common_sp_spec,
    )

    neb_kw = spec1.keywords.copy(update={"maximum_force": 0.051})
    spec2 = spec1.copy(update={"keywords": neb_kw})
    spec_test_runner("neb", spec1, spec2, False)


def test_neb_socket_add_specification_diff_6(spec_test_runner):
    # change align
    spec1 = NEBSpecification(
        program="geometric",
        keywords=NEBKeywords(
            images=21,
            spring_constant=1.5,
            optimize_endpoints=False,
            spring_type=0,
        ),
        singlepoint_specification=common_sp_spec,
    )

    neb_kw = spec1.keywords.copy(update={"align": False})
    spec2 = spec1.copy(update={"keywords": neb_kw})
    spec_test_runner("neb", spec1, spec2, False)


def test_neb_socket_add_specification_diff_7(spec_test_runner):
    # change sp basis
    spec1 = NEBSpecification(
        program="geometric",
        keywords=NEBKeywords(
            images=21,
            spring_constant=1.5,
            optimize_endpoints=False,
            spring_type=0,
        ),
        singlepoint_specification=common_sp_spec,
        optimization_specification=common_opt_spec,
    )

    sp_spec = spec1.singlepoint_specification.copy(update={"basis": "def2-qzvp"})
    spec2 = spec1.copy(update={"singlepoint_specification": sp_spec})
    spec_test_runner("neb", spec1, spec2, False)


def test_neb_socket_add_specification_diff_8(spec_test_runner):
    # change opt protocols
    spec1 = NEBSpecification(
        program="geometric",
        keywords=NEBKeywords(
            images=21,
            spring_constant=1.5,
            optimize_endpoints=False,
            spring_type=0,
        ),
        singlepoint_specification=common_sp_spec,
        optimization_specification=common_opt_spec,
    )

    opt_spec = spec1.optimization_specification.copy(update={"protocols": OptimizationProtocols(trajectory="all")})
    spec2 = spec1.copy(update={"optimization_specification": opt_spec})
    spec_test_runner("neb", spec1, spec2, False)



from typing import List, Tuple

from flask import current_app

from qcfractal.flask_app import storage_socket
from qcfractal.flask_app.api_v1.blueprint import api_v1
from qcfractal.flask_app.wrap_route import wrap_global_route
from qcportal.base_models import CommonBulkGetBody, ProjURLParameters
from qcportal.exceptions import LimitExceededError
from qcportal.molecules import Molecule, MoleculeQueryFilters, MoleculeModifyBody, MoleculeUploadOptions
from qcportal.utils import calculate_limit


@api_v1.route("/molecules/<int:molecule_id>", methods=["GET"])
@wrap_global_route("records", "read")
def get_molecules_v1(molecule_id: int, url_params: ProjURLParameters):
    return storage_socket.molecules.get([molecule_id], url_params.include, url_params.exclude)[0]


@api_v1.route("/molecules/bulkGet", methods=["POST"])
@wrap_global_route("records", "read")
def bulk_get_molecules_v1(body_data: CommonBulkGetBody):
    limit = current_app.config["QCFRACTAL_CONFIG"].api_limits.get_molecules
    if len(body_data.ids) > limit:
        raise LimitExceededError(f"Cannot get {len(body_data.ids)} molecule records - limit is {limit}")

    return storage_socket.molecules.get(
        body_data.ids, body_data.include, body_data.exclude, missing_ok=body_data.missing_ok
    )


@api_v1.route("/molecules/<int:molecule_id>", methods=["DELETE"])
@wrap_global_route("records", "delete")
def delete_molecules_v1(molecule_id: int):
    return storage_socket.molecules.delete([molecule_id])


@api_v1.route("/molecules/bulkDelete", methods=["POST"])
@wrap_global_route("records", "delete")
def bulk_delete_molecules_v1(body_data: List[int]):
    return storage_socket.molecules.delete(body_data)


@api_v1.route("/molecules/<int:molecule_id>", methods=["PATCH"])
@wrap_global_route("records", "modify")
def modify_molecules_v1(molecule_id: int, body_data: MoleculeModifyBody):
    return storage_socket.molecules.modify(
        molecule_id=molecule_id,
        name=body_data.name,
        comment=body_data.comment,
        identifiers=body_data.identifiers,
        overwrite_identifiers=body_data.overwrite_identifiers,
    )


@api_v1.route("/molecules/bulkCreate", methods=["POST"])
@wrap_global_route("records", "add")
def add_molecules_v1(body_data: List[Molecule]):
    limit = current_app.config["QCFRACTAL_CONFIG"].api_limits.add_molecules
    if len(body_data) > limit:
        raise LimitExceededError(f"Cannot add {len(body_data)} molecule records - limit is {limit}")

    return storage_socket.molecules.add(body_data)


allowed_extensions = [
    "xyz",
    "json",
    "zip",
    "tar.gz",
    "tgz",
    "tar.bz2",
    "tbz2",
    "tar.xz",
    "txz",
]


@api_v1.route("/molecules/fromFiles", methods=["POST"])
@wrap_global_route("records", "add", allowed_file_extensions=allowed_extensions)
def add_molecules_from_files_v1(body_data: MoleculeUploadOptions, files: List[Tuple[str, str]]):
    return storage_socket.molecules.add_from_files(files, body_data)


@api_v1.route("/molecules/query", methods=["POST"])
@wrap_global_route("records", "read")
def query_molecules_v1(body_data: MoleculeQueryFilters):
    max_limit = current_app.config["QCFRACTAL_CONFIG"].api_limits.get_molecules
    body_data.limit = calculate_limit(max_limit, body_data.limit)

    return storage_socket.molecules.query(body_data)



from __future__ import annotations

import itertools
from typing import TYPE_CHECKING, Optional

import pytest

from qcarchivetesting import load_molecule_data
from qcfractal.components.record_db_models import BaseRecordORM
from qcportal.reaction import ReactionSpecification, ReactionKeywords
from qcportal.record_models import RecordStatusEnum, PriorityEnum
from qcportal.singlepoint import QCSpecification
from qcportal.utils import now_at_utc
from .testing_helpers import compare_reaction_specs, test_specs, run_procedure_data, submit_procedure_data

if TYPE_CHECKING:
    from qcarchivetesting.testing_classes import QCATestingSnowflake
    from qcportal import PortalClient


def test_reaction_client_tag_priority(snowflake_client: PortalClient):
    hooh = load_molecule_data("peroxide2")
    ne4 = load_molecule_data("neon_tetramer")

    for tag, priority in itertools.product(["*", "tag99"], list(PriorityEnum)):
        meta1, id1 = snowflake_client.add_reactions(
            [[(1.0, hooh), (2.0, ne4)]],
            "reaction",
            QCSpecification(
                program="Prog2",
                driver="energy",
                method="Hf",
                basis="def2-TZVP",
                keywords={"tag_priority": [tag, priority]},
            ),
            None,
            ReactionKeywords(),
            compute_tag=tag,
            compute_priority=priority,
        )

        assert meta1.n_inserted == 1

        rec = snowflake_client.get_records(id1, include=["service"])
        assert rec[0].service.compute_tag == tag
        assert rec[0].service.compute_priority == priority


@pytest.mark.parametrize("spec", test_specs)
def test_reaction_client_add_get(
    submitter_client: PortalClient,
    spec: ReactionSpecification,
):
    hooh = load_molecule_data("peroxide2")
    ne4 = load_molecule_data("neon_tetramer")
    water = load_molecule_data("water_dimer_minima")

    time_0 = now_at_utc()
    meta1, id1 = submitter_client.add_reactions(
        [[(1.0, hooh), (2.0, ne4)], [(3.0, hooh), (4.0, water)]],
        spec.program,
        spec.singlepoint_specification,
        spec.optimization_specification,
        spec.keywords,
        compute_tag="tag1",
        compute_priority=PriorityEnum.low,
    )
    time_1 = now_at_utc()
    assert meta1.success

    recs = submitter_client.get_reactions(id1, include=["service", "components"])
    assert len(recs) == 2

    for r in recs:
        assert r.record_type == "reaction"
        assert r.record_type == "reaction"
        assert compare_reaction_specs(spec, r.specification)

        assert r.status == RecordStatusEnum.waiting
        assert r.children_status == {}

        assert r.service.compute_tag == "tag1"
        assert r.service.compute_priority == PriorityEnum.low

        assert r.creator_user == submitter_client.username

        assert time_0 < r.created_on < time_1
        assert time_0 < r.modified_on < time_1

    mol_hash_0 = set(x.molecule.identifiers.molecule_hash for x in recs[0].components)
    mol_hash_1 = set(x.molecule.identifiers.molecule_hash for x in recs[1].components)

    assert mol_hash_0 == {hooh.get_hash(), ne4.get_hash()}
    assert mol_hash_1 == {hooh.get_hash(), water.get_hash()}

    expected_coef = {hooh.get_hash(): 1.0, ne4.get_hash(): 2.0}
    db_coef = {x.molecule.identifiers.molecule_hash: x.coefficient for x in recs[0].components}
    assert expected_coef == db_coef

    expected_coef = {hooh.get_hash(): 3.0, water.get_hash(): 4.0}
    db_coef = {x.molecule.identifiers.molecule_hash: x.coefficient for x in recs[1].components}
    assert expected_coef == db_coef


@pytest.mark.parametrize("spec", test_specs)
@pytest.mark.parametrize("find_existing", [True, False])
def test_reaction_client_add_duplicate(
    submitter_client: PortalClient, spec: ReactionSpecification, find_existing: bool
):
    hooh = load_molecule_data("peroxide2")
    ne4 = load_molecule_data("neon_tetramer")
    water = load_molecule_data("water_dimer_minima")
    all_stoich = [[(1.0, hooh), (2.0, ne4)], [(3.0, hooh), (4.0, water)]]

    meta, id = submitter_client.add_reactions(
        all_stoich,
        spec.program,
        spec.singlepoint_specification,
        spec.optimization_specification,
        spec.keywords,
        compute_tag="tag1",
        compute_priority=PriorityEnum.low,
        find_existing=True,
    )
    assert meta.success
    assert meta.n_inserted == len(all_stoich)

    meta, id2 = submitter_client.add_reactions(
        all_stoich,
        spec.program,
        spec.singlepoint_specification,
        spec.optimization_specification,
        spec.keywords,
        compute_tag="tag1",
        compute_priority=PriorityEnum.low,
        find_existing=find_existing,
    )

    if find_existing:
        assert meta.n_existing == len(all_stoich)
        assert meta.n_inserted == 0
        assert id == id2
    else:
        assert meta.n_existing == 0
        assert meta.n_inserted == len(all_stoich)
        assert set(id).isdisjoint(id2)


def test_reaction_client_add_existing_molecule(snowflake_client: PortalClient):
    spec = test_specs[0]

    hooh = load_molecule_data("peroxide2")
    ne4 = load_molecule_data("neon_tetramer")
    water = load_molecule_data("water_dimer_minima")

    # Add a molecule separately
    _, mol_ids = snowflake_client.add_molecules([hooh])

    # Now add records
    meta1, id1 = snowflake_client.add_reactions(
        [[(1.0, hooh), (2.0, ne4)], [(3.0, hooh), (4.0, water)], [(2.0, ne4), (1.0, hooh)]],
        spec.program,
        spec.singlepoint_specification,
        spec.optimization_specification,
        spec.keywords,
        compute_tag="tag1",
        compute_priority=PriorityEnum.low,
    )

    assert meta1.success

    assert meta1.success
    assert meta1.n_inserted == 2
    assert meta1.n_existing == 1

    recs = snowflake_client.get_reactions(id1, include=["components"])
    assert len(recs) == 3
    assert recs[0].id == recs[2].id
    assert recs[0].id != recs[1].id

    rec_mols_1 = set(x.molecule.id for x in recs[0].components)

    assert mol_ids[0] in rec_mols_1


def test_reaction_client_delete(snowflake: QCATestingSnowflake):
    storage_socket = snowflake.get_storage_socket()
    activated_manager_name, _ = snowflake.activate_manager()
    snowflake_client = snowflake.client()

    rxn_id = run_procedure_data(storage_socket, activated_manager_name, "rxn_H2O_psi4_mp2_optsp")

    with storage_socket.session_scope() as session:
        rec = session.get(BaseRecordORM, rxn_id)
        child_ids = [x.optimization_id for x in rec.components if x.optimization_id is not None]
        child_ids += [x.singlepoint_id for x in rec.components if x.singlepoint_id is not None]

        n_children = len(child_ids) + sum(len(x.optimization_record.trajectory) for x in rec.components)

    meta = snowflake_client.delete_records(rxn_id, soft_delete=True, delete_children=False)
    assert meta.success
    assert meta.deleted_idx == [0]
    assert meta.n_children_deleted == 0

    child_recs = snowflake_client.get_records(child_ids, missing_ok=True)
    assert all(x.status == RecordStatusEnum.complete for x in child_recs)
    rxn_rec = snowflake_client.get_records(rxn_id)
    assert rxn_rec.children_status == {RecordStatusEnum.complete: len(child_ids)}

    snowflake_client.undelete_records(rxn_id)

    meta = snowflake_client.delete_records(rxn_id, soft_delete=True, delete_children=True)
    assert meta.success
    assert meta.deleted_idx == [0]
    assert meta.n_children_deleted == n_children

    child_recs = snowflake_client.get_records(child_ids, missing_ok=True)
    assert all(x.status == RecordStatusEnum.deleted for x in child_recs)
    rxn_rec = snowflake_client.get_records(rxn_id)
    assert rxn_rec.children_status == {RecordStatusEnum.deleted: len(child_ids)}

    meta = snowflake_client.delete_records(rxn_id, soft_delete=False, delete_children=True)
    assert meta.success
    assert meta.deleted_idx == [0]
    assert meta.n_children_deleted == n_children

    recs = snowflake_client.get_reactions(rxn_id, missing_ok=True)
    assert recs is None

    child_recs = snowflake_client.get_records(child_ids, missing_ok=True)
    assert all(x is None for x in child_recs)

    # DB should be pretty empty now
    query_res = snowflake_client.query_records()
    query_res_l = list(query_res)
    assert len(query_res_l) == 0


def test_reaction_client_harddelete_nochildren(snowflake: QCATestingSnowflake):
    storage_socket = snowflake.get_storage_socket()
    activated_manager_name, _ = snowflake.activate_manager()
    snowflake_client = snowflake.client()

    rxn_id = run_procedure_data(storage_socket, activated_manager_name, "rxn_H2O_psi4_mp2_optsp")

    with storage_socket.session_scope() as session:
        rec = session.get(BaseRecordORM, rxn_id)
        child_ids = [x.optimization_id for x in rec.components if x.optimization_id is not None]
        child_ids += [x.singlepoint_id for x in rec.components if x.singlepoint_id is not None]

    meta = snowflake_client.delete_records(rxn_id, soft_delete=False, delete_children=False)
    assert meta.success
    assert meta.deleted_idx == [0]
    assert meta.n_children_deleted == 0

    recs = snowflake_client.get_reactions(rxn_id, missing_ok=True)
    assert recs is None

    child_recs = snowflake_client.get_records(child_ids, missing_ok=True)
    assert all(x is not None for x in child_recs)


def test_reaction_client_delete_opt_inuse(snowflake: QCATestingSnowflake):
    storage_socket = snowflake.get_storage_socket()
    activated_manager_name, _ = snowflake.activate_manager()
    snowflake_client = snowflake.client()

    rxn_id = run_procedure_data(storage_socket, activated_manager_name, "rxn_H2O_psi4_mp2_optsp")

    with storage_socket.session_scope() as session:
        rec = session.get(BaseRecordORM, rxn_id)
        child_ids = [x.optimization_id for x in rec.components if x.optimization_id is not None]
        child_ids += [x.singlepoint_id for x in rec.components if x.singlepoint_id is not None]

    meta = snowflake_client.delete_records(child_ids[0], soft_delete=False)
    assert meta.success is False
    assert meta.error_idx == [0]

    ch_rec = snowflake_client.get_records(child_ids[0])
    assert ch_rec is not None


def test_reaction_client_query(snowflake: QCATestingSnowflake):
    storage_socket = snowflake.get_storage_socket()
    snowflake_client = snowflake.client()

    id_1, _ = submit_procedure_data(storage_socket, "rxn_H2O_psi4_b3lyp_sp")
    id_2, _ = submit_procedure_data(storage_socket, "rxn_H2_psi4_b3lyp_sp")

    query_res = snowflake_client.query_reactions(qc_program=["psi4"])
    query_res_l = list(query_res)
    assert len(query_res_l) == 2

    query_res = snowflake_client.query_reactions(qc_program=["nothing"])
    query_res_l = list(query_res)
    assert len(query_res_l) == 0

    mol_H = load_molecule_data("rxn_H")
    mol_H2 = load_molecule_data("rxn_H2")
    _, init_mol_id = storage_socket.molecules.add([mol_H, mol_H2])

    query_res = snowflake_client.query_reactions(molecule_id=[init_mol_id[0], 9999])
    query_res_l = list(query_res)
    assert len(query_res_l) == 1

    query_res = snowflake_client.query_reactions(molecule_id=[init_mol_id[1], 9999])
    query_res_l = list(query_res)
    assert len(query_res_l) == 2

    # query for basis
    query_res = snowflake_client.query_reactions(qc_basis=["def2-TZvp"])
    query_res_l = list(query_res)
    assert len(query_res_l) == 2

    query_res = snowflake_client.query_reactions(qc_basis=["sTO-3g"])
    query_res_l = list(query_res)
    assert len(query_res_l) == 0

    query_res = snowflake_client.query_reactions(qc_basis=[None])
    query_res_l = list(query_res)
    assert len(query_res_l) == 0

    query_res = snowflake_client.query_reactions(qc_basis=[""])
    query_res_l = list(query_res)
    assert len(query_res_l) == 0

    # query for qc_method
    query_res = snowflake_client.query_reactions(qc_method=["hf"])
    query_res_l = list(query_res)
    assert len(query_res_l) == 0

    query_res = snowflake_client.query_reactions(qc_method=["b3lyP"])
    query_res_l = list(query_res)
    assert len(query_res_l) == 2

    # Query by default returns everything
    query_res = snowflake_client.query_reactions()
    query_res_l = list(query_res)
    assert len(query_res_l) == 2

    # Query by default (with a limit)
    query_res = snowflake_client.query_reactions(limit=1)
    query_res_l = list(query_res)
    assert len(query_res_l) == 1



"""new manybody code

Revision ID: fd95035b773b
Revises: a5a701dc344d
Create Date: 2024-06-11 16:29:38.468745

"""

import sqlalchemy as sa
from alembic import op
from sqlalchemy.dialects import postgresql

# revision identifiers, used by Alembic.
revision = "fd95035b773b"
down_revision = "a5a701dc344d"
branch_labels = None
depends_on = None


def upgrade():
    # ### commands auto generated by Alembic - please adjust! ###
    op.create_table(
        "manybody_specification",
        sa.Column("id", sa.Integer(), nullable=False),
        sa.Column("program", sa.String(), nullable=False),
        sa.Column("bsse_correction", postgresql.ARRAY(sa.String()), nullable=False),
        sa.Column("keywords", postgresql.JSONB(astext_type=sa.Text()), nullable=False),
        sa.Column("protocols", postgresql.JSONB(astext_type=sa.Text()), nullable=False),
        sa.Column("specification_hash", sa.String(), nullable=False),
        sa.CheckConstraint("program = LOWER(program)", name="ck_manybody_specification_program_lower"),
        sa.PrimaryKeyConstraint("id"),
    )
    op.create_index("ix_manybody_specification_program", "manybody_specification", ["program"], unique=False)
    op.create_table(
        "manybody_specification_levels",
        sa.Column("id", sa.Integer(), nullable=False),
        sa.Column("manybody_specification_id", sa.Integer(), nullable=False),
        sa.Column("level", sa.Integer(), nullable=False),
        sa.Column("singlepoint_specification_id", sa.Integer(), nullable=False),
        sa.ForeignKeyConstraint(
            ["manybody_specification_id"],
            ["manybody_specification.id"],
        ),
        sa.ForeignKeyConstraint(
            ["singlepoint_specification_id"],
            ["qc_specification.id"],
        ),
        sa.PrimaryKeyConstraint("id"),
        sa.UniqueConstraint("manybody_specification_id", "level", name="ux_manybody_specification_levels_unique"),
    )
    op.create_index(
        "ix_manybody_specifications_levels_manybody_specification_id",
        "manybody_specification_levels",
        ["manybody_specification_id"],
        unique=False,
    )
    op.create_index(
        "ix_manybody_specifications_levels_singlepoint_specification_id",
        "manybody_specification_levels",
        ["singlepoint_specification_id"],
        unique=False,
    )
    op.create_table(
        "manybody_dataset",
        sa.Column("id", sa.Integer(), nullable=False),
        sa.ForeignKeyConstraint(["id"], ["base_dataset.id"], ondelete="cascade"),
        sa.PrimaryKeyConstraint("id"),
    )
    op.create_table(
        "manybody_record",
        sa.Column("id", sa.Integer(), nullable=False),
        sa.Column("initial_molecule_id", sa.Integer(), nullable=False),
        sa.Column("specification_id", sa.Integer(), nullable=False),
        sa.ForeignKeyConstraint(["id"], ["base_record.id"], ondelete="cascade"),
        sa.ForeignKeyConstraint(
            ["initial_molecule_id"],
            ["molecule.id"],
        ),
        sa.ForeignKeyConstraint(
            ["specification_id"],
            ["manybody_specification.id"],
        ),
        sa.PrimaryKeyConstraint("id"),
    )
    op.create_table(
        "manybody_cluster",
        sa.Column("id", sa.Integer(), nullable=False),
        sa.Column("manybody_id", sa.Integer(), nullable=True),
        sa.Column("molecule_id", sa.Integer(), nullable=False),
        sa.Column("mc_level", sa.String(), nullable=False),
        sa.Column("fragments", postgresql.ARRAY(sa.Integer()), nullable=False),
        sa.Column("basis", postgresql.ARRAY(sa.Integer()), nullable=False),
        sa.Column("singlepoint_id", sa.Integer(), nullable=True),
        sa.CheckConstraint("array_length(basis, 1) > 0", name="ck_manybody_cluster_basis"),
        sa.CheckConstraint("array_length(fragments, 1) > 0", name="ck_manybody_cluster_fragments"),
        sa.ForeignKeyConstraint(["manybody_id"], ["manybody_record.id"], ondelete="cascade"),
        sa.ForeignKeyConstraint(
            ["molecule_id"],
            ["molecule.id"],
        ),
        sa.ForeignKeyConstraint(
            ["singlepoint_id"],
            ["singlepoint_record.id"],
        ),
        sa.PrimaryKeyConstraint("id"),
        sa.UniqueConstraint("manybody_id", "mc_level", "fragments", "basis", name="ux_manybody_cluster_unique"),
    )
    op.create_index("ix_manybody_cluster_molecule_id", "manybody_cluster", ["molecule_id"], unique=False)
    op.create_index("ix_manybody_cluster_singlepoint_id", "manybody_cluster", ["singlepoint_id"], unique=False)
    op.create_table(
        "manybody_dataset_entry",
        sa.Column("dataset_id", sa.Integer(), nullable=False),
        sa.Column("name", sa.String(), nullable=False),
        sa.Column("comment", sa.String(), nullable=True),
        sa.Column("initial_molecule_id", sa.Integer(), nullable=False),
        sa.Column("additional_singlepoint_keywords", postgresql.JSONB(astext_type=sa.Text()), nullable=False),
        sa.Column("attributes", postgresql.JSONB(astext_type=sa.Text()), nullable=False),
        sa.ForeignKeyConstraint(["dataset_id"], ["manybody_dataset.id"], ondelete="cascade"),
        sa.ForeignKeyConstraint(
            ["initial_molecule_id"],
            ["molecule.id"],
        ),
        sa.PrimaryKeyConstraint("dataset_id", "name"),
    )
    op.create_index("ix_manybody_dataset_entry_dataset_id", "manybody_dataset_entry", ["dataset_id"], unique=False)
    op.create_index(
        "ix_manybody_dataset_entry_initial_molecule_id", "manybody_dataset_entry", ["initial_molecule_id"], unique=False
    )
    op.create_index("ix_manybody_dataset_entry_name", "manybody_dataset_entry", ["name"], unique=False)
    op.create_table(
        "manybody_dataset_specification",
        sa.Column("dataset_id", sa.Integer(), nullable=False),
        sa.Column("name", sa.String(), nullable=False),
        sa.Column("description", sa.String(), nullable=True),
        sa.Column("specification_id", sa.Integer(), nullable=False),
        sa.ForeignKeyConstraint(["dataset_id"], ["manybody_dataset.id"], ondelete="cascade"),
        sa.ForeignKeyConstraint(
            ["specification_id"],
            ["manybody_specification.id"],
        ),
        sa.PrimaryKeyConstraint("dataset_id", "name"),
    )
    op.create_index(
        "ix_manybody_dataset_specification_dataset_id", "manybody_dataset_specification", ["dataset_id"], unique=False
    )
    op.create_index("ix_manybody_dataset_specification_name", "manybody_dataset_specification", ["name"], unique=False)
    op.create_index(
        "ix_manybody_dataset_specification_specification_id",
        "manybody_dataset_specification",
        ["specification_id"],
        unique=False,
    )
    op.create_table(
        "manybody_dataset_record",
        sa.Column("dataset_id", sa.Integer(), nullable=False),
        sa.Column("entry_name", sa.String(), nullable=False),
        sa.Column("specification_name", sa.String(), nullable=False),
        sa.Column("record_id", sa.Integer(), nullable=False),
        sa.ForeignKeyConstraint(
            ["dataset_id", "entry_name"],
            ["manybody_dataset_entry.dataset_id", "manybody_dataset_entry.name"],
            onupdate="cascade",
            ondelete="cascade",
        ),
        sa.ForeignKeyConstraint(
            ["dataset_id", "specification_name"],
            ["manybody_dataset_specification.dataset_id", "manybody_dataset_specification.name"],
            onupdate="cascade",
            ondelete="cascade",
        ),
        sa.ForeignKeyConstraint(["dataset_id"], ["manybody_dataset.id"], ondelete="cascade"),
        sa.ForeignKeyConstraint(
            ["record_id"],
            ["manybody_record.id"],
        ),
        sa.PrimaryKeyConstraint("dataset_id", "entry_name", "specification_name"),
    )
    op.create_index("ix_manybody_dataset_record_record_id", "manybody_dataset_record", ["record_id"], unique=False)

    op.execute(
        """CREATE TRIGGER qca_manybody_record_delete_base_tr
                  AFTER DELETE ON public.manybody_record
                  FOR EACH ROW EXECUTE FUNCTION qca_base_record_delete();"""
    )

    # ### end Alembic commands ###


def downgrade():
    # ### commands auto generated by Alembic - please adjust! ###
    op.drop_index("ix_manybody_dataset_record_record_id", table_name="manybody_dataset_record")
    op.drop_table("manybody_dataset_record")
    op.drop_index("ix_manybody_dataset_specification_specification_id", table_name="manybody_dataset_specification")
    op.drop_index("ix_manybody_dataset_specification_name", table_name="manybody_dataset_specification")
    op.drop_index("ix_manybody_dataset_specification_dataset_id", table_name="manybody_dataset_specification")
    op.drop_table("manybody_dataset_specification")
    op.drop_index("ix_manybody_dataset_entry_name", table_name="manybody_dataset_entry")
    op.drop_index("ix_manybody_dataset_entry_initial_molecule_id", table_name="manybody_dataset_entry")
    op.drop_index("ix_manybody_dataset_entry_dataset_id", table_name="manybody_dataset_entry")
    op.drop_table("manybody_dataset_entry")
    op.drop_index("ix_manybody_cluster_singlepoint_id", table_name="manybody_cluster")
    op.drop_index("ix_manybody_cluster_molecule_id", table_name="manybody_cluster")
    op.drop_table("manybody_cluster")
    op.drop_table("manybody_record")
    op.drop_table("manybody_dataset")
    op.drop_index(
        "ix_manybody_specifications_levels_singlepoint_specification_id", table_name="manybody_specification_levels"
    )
    op.drop_index(
        "ix_manybody_specifications_levels_manybody_specification_id", table_name="manybody_specification_levels"
    )
    op.drop_table("manybody_specification_levels")
    op.drop_index("ix_manybody_specification_program", table_name="manybody_specification")
    op.drop_table("manybody_specification")
    # ### end Alembic commands ###



"""add available column on task queue

Revision ID: 12e2ba353ee6
Revises: 73b4838a6839
Create Date: 2024-10-18 10:31:43.514061

"""

import sqlalchemy as sa
from alembic import op

# revision identifiers, used by Alembic.
revision = "12e2ba353ee6"
down_revision = "73b4838a6839"
branch_labels = None
depends_on = None


def upgrade():
    # ### commands auto generated by Alembic - please adjust! ###
    op.add_column("task_queue", sa.Column("available", sa.Boolean(), nullable=True))

    op.execute(
        """
            UPDATE task_queue tq
            SET available = CASE WHEN br.status = 'waiting' THEN TRUE ELSE FALSE END
            FROM base_record br
            WHERE tq.record_id = br.id;
    """
    )

    op.drop_index("ix_task_queue_sort", table_name="task_queue")
    op.execute(
        "CREATE INDEX ix_task_queue_sort ON task_queue (priority DESC, sort_date, id, tag) WHERE available = True;"
    )

    op.alter_column("task_queue", "available", nullable=False)
    # ### end Alembic commands ###


def downgrade():
    # ### commands auto generated by Alembic - please adjust! ###
    op.drop_index("ix_task_queue_sort", table_name="task_queue")
    op.execute("CREATE INDEX ix_task_queue_sort ON task_queue (priority DESC, sort_date, id, tag);")

    op.drop_column("task_queue", "available")
    # ### end Alembic commands ###



from __future__ import annotations

from typing import TYPE_CHECKING, Optional

import pytest

from qcportal.record_models import PriorityEnum

if TYPE_CHECKING:
    from qcarchivetesting.testing_classes import QCATestingSnowflake


@pytest.mark.parametrize(
    "default_tag,default_priority,default_user,default_group",
    [
        ("*", PriorityEnum.low, None, None),
        ("a_tag", PriorityEnum.high, "admin_user", "group1"),
        ("TAG2", PriorityEnum.normal, "submit_user", None),
    ],
)
def test_dataset_socket_submit_defaults(
    secure_snowflake: QCATestingSnowflake,
    default_tag: str,
    default_priority: PriorityEnum,
    default_user: Optional[str],
    default_group: Optional[str],
):
    storage_socket = secure_snowflake.get_storage_socket()

    default_user_id = storage_socket.users.get_optional_user_id(default_user)
    group1_id = storage_socket.groups.get("group1")["id"]

    ds_id = storage_socket.datasets.singlepoint.add(
        name="Test SP Dataset",
        description="",
        tagline="",
        tags=[],
        provenance={},
        default_compute_tag=default_tag,
        default_compute_priority=default_priority,
        extras={},
        creator_user=default_user,
        existing_ok=False,
    )

    ds = storage_socket.datasets.get(ds_id)
    assert ds["default_tag"] == default_tag.lower()
    assert ds["default_priority"] == default_priority
    assert ds["owner_user"] == default_user

    tag, priority = storage_socket.datasets.singlepoint.get_submit_defaults(ds_id)
    assert tag == default_tag.lower()
    assert priority == default_priority

    tag, priority, user_id = storage_socket.datasets.singlepoint.get_submit_info(
        ds_id,
        None,
        None,
        None,
    )
    assert tag == default_tag.lower()
    assert priority == default_priority
    assert user_id is None

    tag, priority, user_id = storage_socket.datasets.singlepoint.get_submit_info(ds_id, None, None, default_user)
    assert tag == default_tag.lower()
    assert priority == default_priority
    assert user_id == default_user_id

    tag, priority, user_id = storage_socket.datasets.singlepoint.get_submit_info(ds_id, None, None, "submit_user")
    assert tag == default_tag.lower()
    assert priority == default_priority

    # No user = no group either
    tag, priority, user_id = storage_socket.datasets.singlepoint.get_submit_info(ds_id, "diFFerent_TAG", None, None)
    assert tag == "different_tag"
    assert priority == default_priority
    assert user_id is None

    tag, priority, user_id = storage_socket.datasets.singlepoint.get_submit_info(
        ds_id, "diFFerent_TAG", None, default_user
    )
    assert tag == "different_tag"
    assert priority == default_priority
    assert user_id == default_user_id

    tag, priority, user_id = storage_socket.datasets.singlepoint.get_submit_info(
        ds_id, "different_tag", PriorityEnum.normal, default_user
    )
    assert tag == "different_tag"
    assert priority == PriorityEnum.normal
    assert user_id == default_user_id
